package ru.dbotthepony.kstarbound.lua;

import com.kenai.jffi.Closure;
import jnr.ffi.LibraryLoader;
import jnr.ffi.Pointer;
import jnr.ffi.Runtime;
import jnr.ffi.annotations.IgnoreError;
import jnr.ffi.annotations.LongLong;
import org.jetbrains.annotations.NotNull;

import javax.annotation.Nullable;

@SuppressWarnings({"UnnecessaryModifier", "SpellCheckingInspection", "unused"})
public interface LuaJNR {
	/**
	 *  Calls a function in protected mode.
	 * <p>
	 * Both nargs and nresults have the same meaning as in lua_call. If there are no errors during the call, lua_pcall behaves exactly like lua_call. However, if there is any error, lua_pcall catches it, pushes a single value on the stack (the error object), and returns an error code. Like lua_call, lua_pcall always removes the function and its arguments from the stack.
	 * <p>
	 * If msgh is 0, then the error object returned on the stack is exactly the original error object. Otherwise, msgh is the stack index of a message handler. (This index cannot be a pseudo-index.) In case of runtime errors, this function will be called with the error object and its return value will be the object returned on the stack by lua_pcall.
	 * <p>
	 * Typically, the message handler is used to add more debug information to the error object, such as a stack traceback. Such information cannot be gathered after the return of lua_pcall, since by then the stack has unwound.
	 * <p>
	 * The lua_pcall function returns one of the following constants (defined in lua.h):
	 * <p>
	 *     LUA_OK (0): success.
	 *     LUA_ERRRUN: a runtime error.
	 *     LUA_ERRMEM: memory allocation error. For such errors, Lua does not call the message handler.
	 *     LUA_ERRERR: error while running the message handler.
	 *     LUA_ERRGCMM: error while running a __gc metamethod. For such errors, Lua does not call the message handler (as this kind of error typically has no relation with the function being called).
	 */
	@IgnoreError
	public int lua_pcallk(@NotNull Pointer luaState, int numArgs, int numResults, int msgh, @LongLong long ctx, @LongLong long callback);
	@IgnoreError
	public int lua_callk(@NotNull Pointer luaState, int numArgs, int numResults, @LongLong long ctx, @LongLong long callback);
	@IgnoreError
	public long lua_atpanic(@NotNull Pointer luaState, @LongLong long fn);

	/**
	 * Creates a new thread, pushes it on the stack, and returns a pointer to a lua_State that represents this new thread. The new thread returned by this function shares with the original thread its global environment, but has an independent execution stack.
	 * <p>
	 * There is no explicit function to close or to destroy a thread. Threads are subject to garbage collection, like any Lua object.
	 */
	@NotNull
	@IgnoreError
	public Pointer lua_newthread(@NotNull Pointer luaState);

	@IgnoreError
	public int luaL_ref(@NotNull Pointer luaState, int stackIndex);
	@IgnoreError
	public void lua_copy(@NotNull Pointer luaState, int fromIndex, int toIndex);
	@IgnoreError
	public int lua_xmove(@NotNull Pointer from, @NotNull Pointer to, int amount);

	@Nullable
	public Pointer luaL_newstate();
	public void lua_close(@NotNull Pointer luaState);

	// Стандартные библиотеки
	@IgnoreError
	public void luaopen_base(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_package(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_coroutine(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_table(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_io(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_os(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_string(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_math(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_utf8(@NotNull Pointer luaState);
	@IgnoreError
	public void luaopen_debug(@NotNull Pointer luaState);

	@Nullable
	@IgnoreError
	public Pointer lua_tolstring(@NotNull Pointer luaState, int index, @LongLong long statusCodeReturnPtr);

	@IgnoreError
	public int lua_load(@NotNull Pointer luaState, @LongLong long reader, long userData, @NotNull String chunkName, @NotNull String mode);

	/**
	 * Pops a table from the stack and sets it as the new metatable for the value at the given index.
	 */
	@IgnoreError
	public void lua_setmetatable(@NotNull Pointer luaState, int stackIndex);

	/**
	 * Pushes onto the stack the field e from the metatable of the object at index obj and returns the type of the pushed value. If the object does not have a metatable, or if the metatable does not have this field, pushes nothing and returns LUA_TNIL.
	 */
	@IgnoreError
	public int luaL_getmetafield(@NotNull Pointer luaState, int stackIndex, @LongLong long stringPtr);

	public interface lua_CFunction extends Closure {
		int invoke(@NotNull Pointer luaState);

		@Override
		default void invoke(Buffer buffer) {
			//buffer.setIntReturn(invoke());
		}
	}

	public interface lua_Reader extends Closure {
		public long readNextChunk(@NotNull Pointer luaState, long userData, @NotNull Pointer sizeToRead);

		@Override
		default void invoke(Buffer buffer) {
			throw new RuntimeException();
		}
	}

	public interface lua_KFunction extends Closure {
		public int invoke(@NotNull Pointer luaState, int status, long context);

		@Override
		default void invoke(Buffer buffer) {
			throw new RuntimeException();
		}
	}

	// Операции над стаком
	// загрузка значений из Java на стек
	@IgnoreError
	public void lua_createtable(@NotNull Pointer luaState, int arraySize, int hashSize);
	@IgnoreError
	public void lua_pushnil(@NotNull Pointer luaState);
	@IgnoreError
	public void lua_pushnumber(@NotNull Pointer luaState, double value);
	@IgnoreError
	public void lua_pushinteger(@NotNull Pointer luaState, @LongLong long value);
	@IgnoreError
	public void lua_pushboolean(@NotNull Pointer luaState, int value);
	@IgnoreError
	public void lua_pushglobaltable(@NotNull Pointer luaState);

	// NUL терминированная строка
	@IgnoreError
	public void lua_pushstring(@NotNull Pointer luaState, @NotNull String value);

	// двоичная строка
	@IgnoreError
	public long lua_pushlstring(@NotNull Pointer luaState, @LongLong long stringPointer, @LongLong long length);

	// Загрузка Lua значений на стек
	@IgnoreError
	public int lua_getglobal(@NotNull Pointer luaState, @NotNull String name);

	// запись значений со стека
	@IgnoreError
	public void lua_settable(@NotNull Pointer luaState, int stackIndex);
	@IgnoreError
	public void lua_setglobal(@NotNull Pointer luaState, @NotNull String name);

	// проверка стека
	@IgnoreError
	public boolean lua_checkstack(@NotNull Pointer luaState, int value);
	@IgnoreError
	public int lua_absindex(@NotNull Pointer luaState, int value);
	@IgnoreError
	public int lua_iscfunction(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_isinteger(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_isnumber(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_isstring(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_isuserdata(@NotNull Pointer luaState, int index);

	@IgnoreError
	public int lua_toboolean(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_tocfunction(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_toclose(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_gettable(@NotNull Pointer luaState, int index);

	@LongLong
	@IgnoreError
	public long lua_tointegerx(@NotNull Pointer luaState, int index, @LongLong long successCode);
	@IgnoreError
	public double lua_tonumberx(@NotNull Pointer luaState, int index, @LongLong long successCode);

	@IgnoreError
	public int lua_settop(@NotNull Pointer luaState, int index);

	@IgnoreError
	public int lua_next(@NotNull Pointer luaState, int index);
	@IgnoreError
	public int lua_type(@NotNull Pointer luaState, int index);

	/**
	 * Returns the index of the top element in the stack.
	 * Because indices start at 1, this result is equal to the number of elements in the stack; in particular, 0 means an empty stack.
	 */
	@IgnoreError
	public int lua_gettop(@NotNull Pointer luaState);

	@NotNull
	public static final LuaJNR INSTANCE = LibraryLoader.create(LuaJNR.class).load("lua-5.3.6");

	@NotNull
	public static final Runtime RUNTIME = Runtime.getRuntime(INSTANCE);
}
