package ru.dbotthepony.kstarbound.defs

import com.google.gson.JsonObject
import ru.dbotthepony.kommons.io.readBinaryString
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.actor.NPCVariant
import ru.dbotthepony.kstarbound.defs.monster.MonsterVariant
import ru.dbotthepony.kstarbound.defs.`object`.ObjectType
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.json.readJsonElement
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.ItemDropEntity
import ru.dbotthepony.kstarbound.world.entities.MonsterEntity
import ru.dbotthepony.kstarbound.world.entities.NPCEntity
import ru.dbotthepony.kstarbound.world.entities.ProjectileEntity
import ru.dbotthepony.kstarbound.world.entities.StagehandEntity
import ru.dbotthepony.kstarbound.world.entities.player.PlayerEntity
import ru.dbotthepony.kstarbound.world.entities.tile.ContainerObject
import ru.dbotthepony.kstarbound.world.entities.tile.LoungeableObject
import ru.dbotthepony.kstarbound.world.entities.tile.PlantEntity
import ru.dbotthepony.kstarbound.world.entities.tile.PlantPieceEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import java.io.DataInputStream
import java.util.Collections
import java.util.EnumSet

enum class EntityType(override val jsonName: String, val storeName: String, val canBeCreatedByClient: Boolean, val canBeSpawnedByClient: Boolean, val ephemeralIfSpawnedByClient: Boolean = true) : IStringSerializable {
	PLANT("plant", "PlantEntity", false, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return PlantEntity(stream, isLegacy)
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			return PlantEntity(data)
		}
	},

	OBJECT("object", "ObjectEntity", false, true) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			val name = stream.readInternedString()
			val parameters = stream.readJsonElement()
			val config = Registries.worldObjects.getOrThrow(name)
			return WorldObject.create(config, parameters = parameters as? JsonObject ?: JsonObject())!!
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			val prototype = Registries.worldObjects[data["name"]?.asString ?: throw IllegalArgumentException("Missing object name")] ?: throw IllegalArgumentException("No such object defined for '${data["name"]}'")

			val result = when (prototype.value.objectType) {
				ObjectType.OBJECT -> WorldObject(prototype)
				ObjectType.LOUNGEABLE -> LoungeableObject(prototype)
				ObjectType.CONTAINER -> ContainerObject(prototype)
				ObjectType.FARMABLE -> TODO("ObjectType.FARMABLE")
				ObjectType.TELEPORTER -> TODO("ObjectType.TELEPORTER")
				ObjectType.PHYSICS -> TODO("ObjectType.PHYSICS")
			}

			result.deserialize(data)
			return result
		}
	},

	VEHICLE("vehicle", "VehicleEntity", false, true, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			TODO("VEHICLE")
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			TODO("VEHICLE")
		}
	},

	ITEM_DROP("itemDrop", "ItemDropEntity", false, true, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return ItemDropEntity(stream, isLegacy)
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			return ItemDropEntity(data)
		}
	},

	PLANT_DROP("plantDrop", "PlantDropEntity", false, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return PlantPieceEntity(stream, isLegacy)
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			return PlantPieceEntity(data)
		}
	},

	PROJECTILE("projectile", "ProjectileEntity", true, true) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return ProjectileEntity(Registries.projectiles.getOrThrow(stream.readBinaryString()), stream, isLegacy)
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			throw UnsupportedOperationException("NYI: Projectiles are not persistent")
		}
	},

	STAGEHAND("stagehand", "StagehandEntity", true, true) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return StagehandEntity(stream, isLegacy)
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			return StagehandEntity(data)
		}
	},

	MONSTER("monster", "MonsterEntity", false, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return MonsterEntity(MonsterVariant.read(stream, isLegacy))
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			val variantJson = data["monsterVariant"].asJsonObject

			val variant = if (variantJson.size() == 3) {
				val type = variantJson["type"].asString
				val seed = variantJson["seed"].asLong
				val uniqueParameters = variantJson["uniqueParameters"].asJsonObject
				Registries.monsterTypes.getOrThrow(type).value.create(seed, uniqueParameters)
			} else {
				Starbound.gson.fromJsonFast(variantJson, MonsterVariant::class.java)
			}

			val entity = MonsterEntity(variant)
			entity.deserialize(data)
			return entity
		}
	},

	NPC("npc", "NpcEntity", false, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return NPCEntity(NPCVariant.read(stream, isLegacy))
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			val entity = NPCEntity(Starbound.gson.fromJsonFast(data["npcVariant"], NPCVariant::class.java))
			entity.deserialize(data)
			return entity
		}
	},

	PLAYER("player", "PlayerEntity", true, false) {
		override suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity {
			return PlayerEntity(stream, isLegacy)
		}

		override fun fromStorage(data: JsonObject): AbstractEntity {
			return PlayerEntity(data)
		}
	};

	abstract suspend fun fromNetwork(stream: DataInputStream, isLegacy: Boolean): AbstractEntity
	abstract fun fromStorage(data: JsonObject): AbstractEntity

	companion object {
		val ALL: Set<EntityType> = Collections.unmodifiableSet(EnumSet.allOf(EntityType::class.java))
	}
}
