package ru.dbotthepony.kstarbound.defs.actor

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.TypeAdapter
import com.google.gson.annotations.JsonAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import kotlinx.coroutines.future.await
import kotlinx.coroutines.runBlocking
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.gson.stream
import ru.dbotthepony.kommons.io.readBinaryString
import ru.dbotthepony.kommons.io.readMap
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.io.writeMap
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.defs.EntityDamageTeam
import ru.dbotthepony.kstarbound.defs.TeamType
import ru.dbotthepony.kstarbound.defs.TouchDamage
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.io.readColor
import ru.dbotthepony.kstarbound.io.readDouble
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.io.writeColor
import ru.dbotthepony.kstarbound.io.writeDouble
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.json.readJsonElement
import ru.dbotthepony.kstarbound.json.writeJsonElement
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.util.coalesceNull
import ru.dbotthepony.kstarbound.util.random.nextRange
import ru.dbotthepony.kstarbound.util.random.random
import java.io.DataInputStream
import java.io.DataOutputStream
import kotlin.jvm.optionals.getOrNull
import kotlin.math.max

@JsonAdapter(NPCVariant.Adapter::class)
data class NPCVariant(
	val species: Registry.Entry<Species>,
	val seed: Long,
	val typeName: String,
	val level: Double,
	val overrides: JsonElement,
	val humanoidConfig: HumanoidConfig,
	val humanoidIdentity: HumanoidIdentity,
	val scripts: ImmutableList<AssetPath>,
	val initialScriptDelta: Double = 5.0,
	val scriptConfig: JsonElement = JsonNull.INSTANCE,
	val movementParameters: ActorMovementParameters = ActorMovementParameters.EMPTY,
	val statusControllerSettings: StatusControllerConfig = StatusControllerConfig.EMPTY,
	val innateStatusEffects: ImmutableList<PersistentStatusEffect> = ImmutableList.of(),
	val touchDamage: TouchDamage = TouchDamage.EMPTY,
	val disableWornArmor: Boolean = true,
	val dropPools: ImmutableList<Registry.Ref<TreasurePoolDefinition>> = ImmutableList.of(),
	val persistent: Boolean = false,
	val keepAlive: Boolean = false,
	val damageTeam: Int = 0,
	val damageTeamType: TeamType = TeamType.ENEMY,
	val nametagColor: RGBAColor = RGBAColor.WHITE,
	val items: ImmutableMap<String, ItemDescriptor> = ImmutableMap.of(),
) {
	@JsonFactory
	data class SerializedData(
		val levelVariance: Vector2d = Vector2d.ZERO,
		val scripts: ImmutableList<AssetPath>,
		val initialScriptDelta: Double = 5.0,
		val scriptConfig: JsonElement = JsonNull.INSTANCE,
		val humanoidConfig: String? = null,
		val npcname: String? = null,
		val nameGen: ImmutableList<String>? = null,
		val movementParameters: ActorMovementParameters = ActorMovementParameters.EMPTY,
		val statusControllerSettings: StatusControllerConfig = StatusControllerConfig.EMPTY,
		val innateStatusEffects: ImmutableList<PersistentStatusEffect> = ImmutableList.of(),
		val touchDamage: TouchDamage = TouchDamage.EMPTY,
		val disableWornArmor: Boolean = true,
		val dropPools: ImmutableList<Registry.Ref<TreasurePoolDefinition>> = ImmutableList.of(),
		val persistent: Boolean = false,
		val keepAlive: Boolean = false,
		val damageTeam: Int = 0,
		val damageTeamType: TeamType = TeamType.ENEMY,
		val nametagColor: RGBAColor = RGBAColor.WHITE,
		val matchColorIndices: Boolean = false,
	)

	val team: EntityDamageTeam
		get() = EntityDamageTeam(damageTeamType, damageTeam)

	fun check() {
		if (Starbound.DEBUG_BUILD) {
			humanoidIdentity.check()
		}
	}

	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeBinaryString(species.key)
		stream.writeBinaryString(typeName)
		stream.writeDouble(level, isLegacy)
		stream.writeLong(seed)
		stream.writeJsonElement(overrides)

		if (isLegacy)
			stream.writeInt(initialScriptDelta.toInt())
		else
			stream.writeDouble(initialScriptDelta)

		humanoidIdentity.write(stream, isLegacy)
		stream.writeMap(items, { writeBinaryString(it) }, { it.write(this) })
		stream.writeBoolean(persistent)
		stream.writeBoolean(keepAlive)
		stream.writeByte(damageTeam)
		stream.writeByte(damageTeamType.ordinal)

		if (!isLegacy) {
			stream.writeBoolean(disableWornArmor)
			touchDamage.write(stream)
			stream.writeColor(nametagColor)
		}
	}

	@JsonFactory
	data class DiskSerializedData(
		val species: Registry.Entry<Species>,
		val typeName: String,
		val level: Double,
		val seed: Long,
		val overrides: JsonElement,
		val initialScriptDelta: Double = 5.0,
		val humanoidIdentity: HumanoidIdentity,
		val items: ImmutableMap<String, ItemDescriptor>,
		val persistent: Boolean,
		val keepAlive: Boolean,
		val damageTeam: Int,
		val damageTeamType: TeamType,
		val touchDamage: TouchDamage? = null,
		val disableWornArmor: Boolean? = null,
		val nametagColor: RGBAColor? = null,
	)

	class Adapter(gson: Gson) : TypeAdapter<NPCVariant>() {
		private val parent = gson.getAdapter(DiskSerializedData::class.java)

		override fun write(out: JsonWriter, value: NPCVariant) {
			parent.write(out, DiskSerializedData(
				species = value.species,
				typeName = value.typeName,
				level = value.level,
				seed = value.seed,
				overrides = value.overrides,
				initialScriptDelta = value.initialScriptDelta,
				humanoidIdentity = value.humanoidIdentity,
				items = value.items,
				persistent = value.persistent,
				keepAlive = value.keepAlive,
				damageTeam = value.damageTeam,
				damageTeamType = value.damageTeamType,
				touchDamage = value.touchDamage,
				disableWornArmor = value.disableWornArmor,
				nametagColor = value.nametagColor,
			))
		}

		override fun read(`in`: JsonReader): NPCVariant {
			val (
				species,
				typeName,
				level,
				seed,
				overrides,
				initialScriptDelta,
				humanoidIdentity,
				items,
				persistent,
				keepAlive,
				damageTeam,
				damageTeamType,
				touchDamage,
				disableWornArmor,
				nametagColor,
			) = parent.read(`in`)

			val config = Registries.buildNPCConfig(typeName, overrides)
			val serialized = Starbound.gson.fromJson(config, SerializedData::class.java)

			val humanoidConfig = runBlocking { humanoidConfig(serialized, species) }

			val movementParameters = humanoidConfig
				.movementParameters
				.merge(serialized.movementParameters)

			return NPCVariant(
				species = species,
				seed = seed,
				typeName = typeName,
				level = level,
				humanoidConfig = humanoidConfig,
				humanoidIdentity = humanoidIdentity,
				overrides = overrides.deepCopy(),
				scripts = serialized.scripts,
				initialScriptDelta = initialScriptDelta,
				scriptConfig = serialized.scriptConfig,
				movementParameters = movementParameters,
				statusControllerSettings = serialized.statusControllerSettings,
				innateStatusEffects = ImmutableList.copyOf(innates(serialized, level)),
				touchDamage = touchDamage ?: serialized.touchDamage,
				disableWornArmor = disableWornArmor ?: serialized.disableWornArmor,
				dropPools = serialized.dropPools,
				persistent = persistent,
				keepAlive = keepAlive,
				damageTeam = damageTeam,
				damageTeamType = damageTeamType,
				nametagColor = nametagColor ?: serialized.nametagColor,
				items = ImmutableMap.copyOf(items)
			).also { it.check() }
		}
	}

	companion object {
		suspend fun humanoidConfig(serialized: SerializedData, species: Registry.Entry<Species>): HumanoidConfig {
			if (serialized.humanoidConfig != null)
				return Starbound.gson.fromJson(Starbound.loadJsonAsset(serialized.humanoidConfig).await() as JsonObject, HumanoidConfig::class.java)
			else
				return species.value.humanoidConfig()
		}

		 fun innates(serialized: SerializedData, level: Double): ArrayList<PersistentStatusEffect> {
			val innateStatusEffects = ArrayList(serialized.innateStatusEffects)

			innateStatusEffects.add(
				Either.left(StatModifier.value("powerMultiplier", Registries.jsonFunctions.getOrThrow("npcLevelPowerMultiplierModifier").value.evaluate(level)))
			)

			innateStatusEffects.add(
				Either.left(StatModifier.multBase("protection", Registries.jsonFunctions.getOrThrow("npcLevelProtectionMultiplier").value.evaluate(level)))
			)

			innateStatusEffects.add(
				Either.left(StatModifier.multBase("maxHealth", Registries.jsonFunctions.getOrThrow("npcLevelHealthMultiplier").value.evaluate(level)))
			)

			innateStatusEffects.add(
				Either.left(StatModifier.multBase("maxEnergy", Registries.jsonFunctions.getOrThrow("npcLevelEnergyMultiplier").value.evaluate(level)))
			)

			return innateStatusEffects
		}

		@Suppress("BlockingMethodInNonBlockingContext")
		suspend fun read(stream: DataInputStream, isLegacy: Boolean): NPCVariant {
			val species = Registries.species.getOrThrow(stream.readBinaryString())
			val type = stream.readInternedString()
			val level = stream.readDouble(isLegacy)
			val seed = stream.readLong()
			val overrides = stream.readJsonElement()

			val initialScriptDelta = if (isLegacy) stream.readInt().toDouble() else stream.readDouble()
			val identity = HumanoidIdentity.read(stream, isLegacy)
			val items = stream.readMap({ readInternedString() }, { ItemDescriptor(this) })
			val persistent = stream.readBoolean()
			val keepAlive = stream.readBoolean()
			val damageTeam = stream.readUnsignedByte()
			val damageTeamType = TeamType.entries[stream.readUnsignedByte()]

			val config = Registries.buildNPCConfig(type, overrides)
			val serialized = Starbound.gson.fromJson(config, SerializedData::class.java)

			val humanoidConfig = humanoidConfig(serialized, species)

			val movementParameters = humanoidConfig
				.movementParameters
				.merge(serialized.movementParameters)

			val disableWornArmor = if (isLegacy) serialized.disableWornArmor else stream.readBoolean()
			val touchDamage = if (isLegacy) serialized.touchDamage else TouchDamage(stream)
			val nametagColor = if (isLegacy) serialized.nametagColor else stream.readColor()

			return NPCVariant(
				species = species,
				seed = seed,
				typeName = type,
				level = level,
				humanoidConfig = humanoidConfig,
				humanoidIdentity = identity,
				overrides = overrides.deepCopy(),
				scripts = serialized.scripts,
				initialScriptDelta = initialScriptDelta,
				scriptConfig = serialized.scriptConfig,
				movementParameters = movementParameters,
				statusControllerSettings = serialized.statusControllerSettings,
				innateStatusEffects = ImmutableList.copyOf(innates(serialized, level)),
				touchDamage = touchDamage,
				disableWornArmor = disableWornArmor,
				dropPools = serialized.dropPools,
				persistent = persistent,
				keepAlive = keepAlive,
				damageTeam = damageTeam,
				damageTeamType = damageTeamType,
				nametagColor = nametagColor,
				items = ImmutableMap.copyOf(items)
			).also { it.check() }
		}

		suspend fun create(species: Registry.Entry<Species>, type: String, level: Double, seed: Long, overrides: JsonElement = JsonNull.INSTANCE): NPCVariant {
			val random = random(seed)
			val config = Registries.buildNPCConfig(type, overrides)

			val serialized = Starbound.gson.fromJson(config, SerializedData::class.java)
			val finalLevel = max(0.0, random.nextRange(serialized.levelVariance) + level)

			val humanoidConfig = humanoidConfig(serialized, species)

			var identity = species.value.generateHumanoid(random)

			val name = serialized.npcname ?: serialized.nameGen?.let { Starbound.generateNameAsync(it[identity.gender.ordinal], random) } ?: ""
			val personality = humanoidConfig.personalities.random(random)

			identity = identity.copy(
				name = name,
				personalityIdle = personality.idle,
				personalityArmIdle = personality.armIdle,
				personalityHeadOffset = personality.headOffset,
				personalityArmOffset = personality.armOffset
			)

			if ("identity" in config)
				identity = Starbound.gson.fromJson(mergeJson(Starbound.gson.toJsonTree(identity), config["identity"]!!), HumanoidIdentity::class.java)

			val movementParameters = humanoidConfig
				.movementParameters
				.merge(serialized.movementParameters)

			val items = ImmutableMap.Builder<String, ItemDescriptor>()

			if ("items" in config) {
				val itemsConfig = config.getAsJsonObject("items")

				val speciesItemsConfig = itemsConfig["override"]?.coalesceNull?.asJsonArray
					?: itemsConfig[species.key]?.coalesceNull?.asJsonArray
					?: itemsConfig["default"]?.coalesceNull?.asJsonArray

				if (speciesItemsConfig != null) {
					val highestLevelItemsConfig = speciesItemsConfig.stream()
						.map { it.asJsonArray.let { it[0].asDouble to it[1] } }
						.filter { it.first <= finalLevel }
						.sorted { o1, o2 -> o2.first.compareTo(o1.first) }
						.findFirst()
						.getOrNull()?.second as JsonArray?

					if (highestLevelItemsConfig != null) {
						var randomColorIndex = -1

						for (itemSlotConfig in highestLevelItemsConfig.random(random).asJsonObject.entrySet()) {
							var item = ItemDescriptor(itemSlotConfig.value.asJsonArray.random(random))
							val colorIndex = item.parameters["colorIndex"]

							// Randomize color index if colorIndex is an array
							if (colorIndex is JsonArray) {
								if (!serialized.matchColorIndices || randomColorIndex == -1) {
									randomColorIndex = colorIndex.random(random).asInt
								}

								item = item.applyParameters(JsonObject().also {
									it["colorIndex"] = JsonPrimitive(randomColorIndex)
								})
							}

							items.put(itemSlotConfig.key, item)
						}
					}
				}
			}

			return NPCVariant(
				species = species,
				seed = seed,
				typeName = type,
				level = finalLevel,
				humanoidConfig = humanoidConfig,
				humanoidIdentity = identity,
				overrides = overrides.deepCopy(),
				scripts = serialized.scripts,
				initialScriptDelta = serialized.initialScriptDelta,
				scriptConfig = serialized.scriptConfig,
				movementParameters = movementParameters,
				statusControllerSettings = serialized.statusControllerSettings,
				innateStatusEffects = ImmutableList.copyOf(innates(serialized, finalLevel)),
				touchDamage = serialized.touchDamage,
				disableWornArmor = serialized.disableWornArmor,
				dropPools = serialized.dropPools,
				persistent = serialized.persistent,
				keepAlive = serialized.keepAlive,
				damageTeam = serialized.damageTeam,
				damageTeamType = serialized.damageTeamType,
				nametagColor = serialized.nametagColor,
				items = items.buildOrThrow()
			).also { it.check() }
		}
	}
}
