package ru.dbotthepony.kstarbound.defs.actor

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonObject
import kotlinx.coroutines.future.await
import ru.dbotthepony.kommons.collect.filterNotNull
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.defs.ColorReplacements
import ru.dbotthepony.kstarbound.defs.StatusEffectDefinition
import ru.dbotthepony.kstarbound.defs.image.SpriteReference
import ru.dbotthepony.kstarbound.defs.actor.player.BlueprintLearnList
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.fromJson
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.util.getWrapAround
import ru.dbotthepony.kstarbound.util.random.random
import java.util.random.RandomGenerator

@JsonFactory
data class Species(
	val kind: String,
	@Deprecated("Use humanoidConfig() function", replaceWith = ReplaceWith("this.humanoidConfig()"))
	val humanoidConfig: AssetPath = AssetPath("/humanoid.config"),
	@Deprecated("Use humanoidConfig() function", replaceWith = ReplaceWith("this.humanoidConfig()"))
	val humanoidOverrides: JsonObject = JsonObject(),
	val charCreationTooltip: Tooltip = Tooltip(),
	val nameGen: ImmutableList<String>,
	val ouchNoises: ImmutableList<AssetPath>,
	val charGenTextLabels: ImmutableList<String> = ImmutableList.of(),
	val skull: AssetPath = AssetPath("/humanoid/any/dead.png"), // ваш свет угасает
	val defaultBlueprints: BlueprintLearnList,

	val effectDirectives: String = "",

	val headOptionAsHairColor: Boolean = false,
	val headOptionAsFacialhair: Boolean = false,
	val altOptionAsUndyColor: Boolean = false,
	val altOptionAsHairColor: Boolean = false,
	val altOptionAsFacialMask: Boolean = false,
	val hairColorAsBodySubColor: Boolean = false,
	val bodyColorAsFacialMaskSubColor: Boolean = false,
	val altColorAsFacialMaskSubColor: Boolean = false,

	val bodyColor: ImmutableList<ColorReplacements>,
	val undyColor: ImmutableList<ColorReplacements>,
	val hairColor: ImmutableList<ColorReplacements>,
	val genders: ImmutableList<GenderSettings>,
	val statusEffects: ImmutableSet<Registry.Ref<StatusEffectDefinition>> = ImmutableSet.of(),
) {
	@JsonFactory
	data class Tooltip(val title: String = "", val subTitle: String = "", val description: String = "")

	@JsonFactory
	data class GenderSettings(
		val name: String = "",
		val image: SpriteReference = SpriteReference.NEVER,
		val characterImage: SpriteReference = SpriteReference.NEVER,
		val hairGroup: String = "hair",
		val hair: ImmutableSet<String>,
		val shirt: ImmutableSet<ItemDescriptor>,
		val pants: ImmutableSet<ItemDescriptor>,
		val facialHairGroup: String = "",
		val facialHair: ImmutableSet<String> = ImmutableSet.of(),
		val facialMaskGroup: String = "",
		val facialMask: ImmutableList<String> = ImmutableList.of(),
	)

	suspend fun humanoidConfig(): HumanoidConfig {
		return Starbound.gson.fromJson(mergeJson(Starbound.loadJsonAsset(humanoidConfig.fullPath).thenApply { it ?: JsonObject() }.await().deepCopy(), humanoidOverrides), HumanoidConfig::class.java)
	}

	fun buildStatusEffects(): List<PersistentStatusEffect> {
		return statusEffects.stream().filter { it.isPresent }.map { Either.right<StatModifier, Registry.Ref<StatusEffectDefinition>>(it) }.toList()
	}

	suspend fun generateHumanoid(random: RandomGenerator): HumanoidIdentity {
		val gender = Gender.valueOf(random.nextBoolean())
		val name = Starbound.generateNameAsync(nameGen[gender.ordinal], random)

		val genderSettings = genders[gender.ordinal]
		var bodyColor = bodyColor.random(random).toImageOperator()

		var altColor = ""

		val altOpt = random.nextInt()
		val headOpt = random.nextInt()
		val hairOpt = random.nextInt()

		if (altOptionAsUndyColor) {
			altColor = undyColor.getWrapAround(altOpt).toImageOperator()
		}

		val hair = genderSettings.hair.getWrapAround(hairOpt)
		var hairColor = bodyColor

		if (headOptionAsHairColor) {
			hairColor = this.hairColor.getWrapAround(headOpt).toImageOperator()
			if (altOptionAsHairColor) hairColor += this.undyColor.getWrapAround(altOpt).toImageOperator()
		}

		if (hairColorAsBodySubColor)
			bodyColor += hairColor

		var facialHair = ""
		var facialHairGroup = ""
		var facialHairDirective = ""

		if (headOptionAsFacialhair) {
			facialHair = genderSettings.facialHair.getWrapAround(headOpt)
			facialHairGroup = genderSettings.facialHairGroup
			facialHairDirective = hairColor
		}

		var facialMask = ""
		var facialMaskGroup = ""
		var facialMaskDirectives = ""

		if (altOptionAsFacialMask) {
			facialMask = genderSettings.facialMask.getWrapAround(altOpt)
			facialMaskGroup = genderSettings.facialMaskGroup
		}

		if (bodyColorAsFacialMaskSubColor)
			facialMaskDirectives += bodyColor

		if (altColorAsFacialMaskSubColor)
			facialMaskDirectives += altColor

		return HumanoidIdentity(
			name = name,
			species = Registries.species.getOrThrow(kind).ref,
			gender = gender,
			hairGroup = genderSettings.hairGroup,
			hairType = hair,
			hairDirectives = hairColor,
			bodyDirectives = bodyColor + altColor,
			emoteDirectives = bodyColor + altColor,
			facialHairGroup = facialHairGroup,
			facialHairType = facialHair,
			facialHairDirectives = facialHairDirective,
			facialMaskGroup = facialMaskGroup,
			facialMaskType = facialMask,
			facialMaskDirectives = facialMaskDirectives
		).also { it.check() }
	}
}
