package ru.dbotthepony.kstarbound.defs.monster

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import ru.dbotthepony.kommons.gson.JsonArrayCollector
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.gson.getArray
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.AssetReference
import ru.dbotthepony.kstarbound.defs.animation.AnimationDefinition
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.random.MWCRandom
import ru.dbotthepony.kstarbound.util.random.nextUInt
import ru.dbotthepony.kstarbound.util.random.random
import java.util.concurrent.CompletableFuture
import kotlin.math.max
import kotlin.math.min
import kotlin.math.pow

private val specialMerge = ImmutableSet.of("scripts", "skills", "specialSkills", "baseSkills")

private fun mergeFinalParameters(params: List<JsonElement>): JsonObject {
	val parameters = JsonObject()

	for (baseParameters in params) {
		if (baseParameters !is JsonObject)
			continue

		for ((k, v) in baseParameters.entrySet()) {
			// Hard-coded merge for scripts and skills parameters, otherwise merge.
			if (k in specialMerge) {
				v as JsonArray
				val v2 = parameters[k]

				if (v2 !is JsonArray) {
					parameters[k] = v.deepCopy()
				} else {
					val result = JsonArray(v.size() + v2.size())
					result.addAll(v2)
					result.addAll(v)
					parameters[k] = result
				}
			} else {
				parameters[k] = mergeJson((parameters[k] ?: JsonNull.INSTANCE).deepCopy(), v)
			}
		}
	}

	return parameters
}

@JsonFactory
data class MonsterTypeDefinition(
	val type: String,
	val shortdescription: String? = null,
	val description: String? = null,
	val categories: ImmutableSet<String>,
	val parts: ImmutableSet<String>,
	val animation: AssetReference<AnimationDefinition>,
	val colors: String = "default",
	val reversed: Boolean = false,
	val dropPools: ImmutableList<Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>> = ImmutableList.of(),
	val baseParameters: JsonElement = JsonNull.INSTANCE,

	@Deprecated("Raw property, use processed one", replaceWith = ReplaceWith("this.paramsOverrides"))
	val partParameters: JsonElement = JsonNull.INSTANCE,

	@Deprecated("Raw property, use processed one", replaceWith = ReplaceWith("this.paramsDesc"))
	val partParameterDescription: JsonElement = JsonObject()
) {
	val paramsDesc: CompletableFuture<JsonObject>
	val paramsOverrides: CompletableFuture<JsonObject>

	init {
		if (partParameters.isJsonNull) {
			// outdated monsters still have partParameterDescription defined
			// directly in the
			// .monstertype file
			paramsDesc = CompletableFuture.completedFuture(partParameterDescription.asJsonObject)
			paramsOverrides = CompletableFuture.completedFuture(JsonObject())
		} else {
			// for updated monsters, use the partParameterDescription from the
			// .partparams file
			val json = Starbound.loadJsonAsset(partParameters, AssetPathStack.lastFolder())

			paramsDesc = json.thenApply {
				it.asJsonObject.get("partParameterDescription").asJsonObject
			}

			paramsOverrides = json.thenApply {
				it.asJsonObject.get("partParameters").asJsonObject
			}
		}
	}

	fun create(seed: Long, uniqueParameters: JsonObject): MonsterVariant {
		// MWCRandom since this MUST product exactly the same result
		// between legacy client and new server (or new client and legacy server)
		// Thanks Chuckleclusterfuck
		val random = MWCRandom(seed.toULong(), 256, 32)

		val copyParameters = uniqueParameters.deepCopy()

		// select a list of monster parts
		val monsterParts = ArrayList<MonsterPartDefinition>()
		val categoryName = categories.random(random, true)
		var selectedParts = uniqueParameters["selectedParts"]

		// key -> image
		val animatorPartTags = HashMap<String, String>()

		if (selectedParts !is JsonObject) {
			selectedParts = JsonObject()
			uniqueParameters["selectedParts"] = selectedParts
		}

		for (partTypeName in parts) {
			val randPart = Registries.selectMonsterPart(categoryName, partTypeName, random)
			val selectedPart = selectedParts[partTypeName]?.asString

			if (selectedPart != null) {
				val get = Registries.getMonsterPart(categoryName, partTypeName, selectedPart)

				if (get != null) {
					monsterParts.add(get)
				}
			} else if (randPart != null) {
				monsterParts.add(randPart)

				// cheat the system; while this will surely poison uniqueParameters, at least
				// we will make sure selected parts are chosen deterministically across sessions / sides
				// However, this hack will only work if we are server in both meanings;
				//  e.g. we are fully in charge of created monster. If monster is spawned by player,
				//  then, well, it sucks.
				selectedParts[partTypeName] = randPart.name
			}
		}

		for (partConfig in monsterParts) {
			for ((k, v) in partConfig.frames) {
				animatorPartTags[k] = v.fullPath
			}
		}

		val partParameterList = ArrayList<JsonObject>()

		for (partConfig in monsterParts) {
			partParameterList.add(partConfig.parameters)

			// Include part parameter overrides
			if (partConfig.name in paramsOverrides.get()) {
				partParameterList.add(paramsOverrides.get()[partConfig.name].asJsonObject)
			}
		}

		// merge part parameters and unique parameters into base parameters
		var parameters = JsonObject()

		// First assign all the defaults.
		for ((k, v) in paramsDesc.get().entrySet()) {
			parameters[k] = v.asJsonArray.get(1).deepCopy()
		}

		// Then go through parameter list and merge based on the merge rules.
		for (applyParams in partParameterList) {
			for ((k, v) in applyParams.entrySet()) {
				val mergeMethod = paramsDesc.get()[k]?.asJsonArray?.get(0)?.asString ?: "override"
				var value = parameters[k] ?: JsonNull.INSTANCE

				when (mergeMethod.lowercase()) {
					"add" -> value = JsonPrimitive(value.asDouble + v.asDouble)
					"sub" -> value = JsonPrimitive(value.asDouble - v.asDouble)
					"multiply" -> value = JsonPrimitive(value.asDouble * v.asDouble)
					"divide" -> value = JsonPrimitive(value.asDouble / v.asDouble)
					"inverse" -> value = JsonPrimitive(v.asDouble / value.asDouble)
					"min" -> value = JsonPrimitive(min(value.asDouble, v.asDouble))
					"max" -> value = JsonPrimitive(max(value.asDouble, v.asDouble))
					"pow" -> value = JsonPrimitive(value.asDouble.pow(v.asDouble))
					"invpow" -> value = JsonPrimitive(v.asDouble.pow(value.asDouble))

					"override" -> {
						if (!v.isJsonNull) // why does original engine check this?
							value = v
					}

					"merge" -> {
						// "merge" means to either merge maps, or *append* lists together
						if (!v.isJsonNull) {
							if (value.isJsonNull) {
								value = v.deepCopy()
							} else if (value::class != v::class) {
								value = v.deepCopy()
							} else {
								if (v is JsonArray) {
									value = value.asJsonArray.also { it.addAll(v) }
								} else if (v is JsonObject) {
									value = mergeJson(value, v)
								}
							}
						}
					}
				}

				parameters[k] = value
			}
		}

		parameters = mergeFinalParameters(listOf(baseParameters, parameters))
		mergeJson(parameters, uniqueParameters)

		var animationConfig = animation.json.get() ?: JsonNull.INSTANCE
		val skillNames = ArrayList<String>()

		if ("baseSkills" in parameters || "specialSkills" in parameters) {
			val skillCount = parameters.get("skillCount", 2)

			val baseSkillNames = parameters.getArray("baseSkills").deepCopy()
			val specialSkillNames = parameters.getArray("specialSkills").deepCopy()

			// First, pick from base skills
			while (baseSkillNames.size() > 0 && skillNames.size < skillCount) {
				val pick = random.nextUInt(baseSkillNames.size().toULong() - 1UL).toInt()
				skillNames.add(baseSkillNames.remove(pick).asString)
			}

			// ...then fill in from special skills as needed
			while (specialSkillNames.size() > 0 && skillNames.size < skillCount) {
				val pick = random.nextUInt(specialSkillNames.size().toULong() - 1UL).toInt()
				skillNames.add(specialSkillNames.remove(pick).asString)
			}
		} else if ("skills" in parameters) {
			val availableSkillNames = parameters.getArray("skills").deepCopy()
			val skillCount = min(parameters.get("skillCount", 2), availableSkillNames.size())

			while (skillNames.size < skillCount) {
				val pick = random.nextUInt(availableSkillNames.size().toULong() - 1UL).toInt()
				skillNames.add(availableSkillNames.remove(pick).asString)
			}
		}

		if (skillNames.isNotEmpty()) {
			animationConfig = animationConfig.deepCopy()
			val allParameters = ArrayList<JsonElement>()
			allParameters.add(parameters)

			for (skillName in skillNames) {
				val skill = Registries.monsterSkills[skillName] ?: continue
				allParameters.add(skill.value.parameters)
				animationConfig = mergeJson(animationConfig, skill.value.animationParameters)
			}

			// Need to override the final list of skills, instead of merging the lists
			parameters = mergeFinalParameters(allParameters)
			parameters["skills"] = skillNames.stream().map { JsonPrimitive(it) }.collect(JsonArrayCollector)
		}

		val variant = MonsterVariant(
			type = type,
			description = description,
			shortDescription = shortdescription,
			seed = seed,
			uniqueParameters = copyParameters,
			animationConfig = Starbound.gson.fromJsonFast(animationConfig, AnimationDefinition::class.java),
			reversed = reversed,
			animatorPartTags = ImmutableMap.copyOf(animatorPartTags),
			parameters = parameters,
			dropPools = dropPools,
		)

		return variant
	}
}
