package ru.dbotthepony.kstarbound.lua.bindings

import com.google.common.collect.ImmutableMap
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonNull
import com.google.gson.reflect.TypeToken
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.defs.EntityDamageTeam
import ru.dbotthepony.kstarbound.defs.PhysicsForceRegion
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.util.SBPattern
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.entities.ActorEntity
import ru.dbotthepony.kstarbound.world.entities.MonsterEntity

private object DropPoolToken : TypeToken<Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>>()

private fun type(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.type)
	return 1
}

private fun seed(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	// FIXME: what the fuck.
	args.lua.push(self.variant.seed.toString())
	return 1
}

private fun seedNumber(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.seed)
	return 1
}

private fun uniqueParameters(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.uniqueParameters)
	return 1
}

private fun level(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	// FIXME: this makes half sense
	args.lua.push(self.monsterLevel ?: 0.0)
	return 1
}

private fun setDamageOnTouch(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.damageOnTouch = args.nextBoolean()
	return 0
}

private fun setDamageSources(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.customDamageSources.clear()

	if (args.isNotEmpty) {
		val sources = args.readTableValues { Starbound.gson.fromJsonFast(getJson(it)!!, DamageSource::class.java) }
		self.customDamageSources.addAll(sources)
	}

	return 0
}

private fun setDamageParts(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	if (args.isEmpty) {
		self.animationDamageParts.clear()
	} else {
		val strings = ImmutableSet.copyOf(args.readTableValues { getString(it) ?: throw IllegalArgumentException("Invalid values in damage parts table") })
		self.animationDamageParts.removeIf { it !in strings }

		for (v in strings) {
			if (v !in self.animationDamageParts) {
				self.animationDamageParts.add(v)
			}
		}
	}

	return 0
}

private fun setAggressive(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.isAggressive = args.nextBoolean()
	return 0
}

private fun setActiveSkillName(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.activeSkillName = args.nextString().sbIntern()
	return 0
}

private fun setDropPool(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.dropPool = Starbound.gson.fromJsonFast(args.nextOptionalJson() ?: JsonNull.INSTANCE, DropPoolToken)
	return 0
}

private fun toAbsolutePosition(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.movement.getAbsolutePosition(args.nextVector2d()))
	return 0
}

private fun mouthPosition(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.mouthPosition)
	return 0
}

// This callback is registered here rather than in
// makeActorMovementControllerCallbacks
// because it requires access to world
private fun flyTo(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.movement.controlFly = self.world.geometry.diff(args.nextVector2d(), self.movement.position)
	return 0
}

private fun setDeathParticleBurst(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.deathParticlesBurst = args.nextOptionalString()?.sbIntern() ?: ""
	return 0
}

private fun setDeathSound(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.deathSound = args.nextOptionalString()?.sbIntern() ?: ""
	return 0
}

private fun setPhysicsForces(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.forceRegions.clear()

	if (args.isNotEmpty) {
		val forces = args.readTableValues { Starbound.gson.fromJsonFast(getJson(it)!!, PhysicsForceRegion::class.java) }
		self.forceRegions.addAll(forces)
	}

	return 0
}

private fun setName(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.networkName = args.nextOptionalString()?.sbIntern() ?: ""
	return 0
}

private fun setDisplayNametag(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.displayNameTag = args.nextBoolean()
	return 0
}

private fun say(self: MonsterEntity, hasPortrait: Boolean, args: LuaThread.ArgStack): Int {
	var line = args.nextString()
	val portrait = if (hasPortrait) args.nextString() else null

	if (args.hasNext) {
		val tags = args.readTable(keyVisitor = { getString(it) ?: throw IllegalArgumentException("Pattern tag replacement table contains non-string keys") }, valueVisitor = { getString(it) ?: throw IllegalArgumentException("Pattern tag replacement table contains non-string values") })
		line = SBPattern.of(line).resolveOrSkip({ t -> tags.firstOrNull { it.first == t }?.second })
	}

	if (line.isNotBlank()) {
		self.addChatMessage(line.sbIntern(), portrait = portrait)
		args.lua.push(true)
	} else {
		args.lua.push(false)
	}

	return 1
}

private fun setDamageTeam(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.team.accept(Starbound.gson.fromJsonFast(args.nextJson(), EntityDamageTeam::class.java))
	return 0
}

private fun setUniqueId(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.uniqueID.accept(args.nextOptionalString()?.sbIntern())
	return 0
}

private fun setDamageBar(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.damageBarType = ActorEntity.DamageBarType.entries.valueOf(args.nextString())
	return 0
}

private fun setInteractive(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.isInteractive = args.nextBoolean()
	return 0
}

private fun setAnimationParameter(self: MonsterEntity, args: LuaThread.ArgStack): Int {
	self.scriptedAnimationParameters[args.nextString().sbIntern()] = args.nextJson()
	return 0
}

fun provideMonsterBindings(self: MonsterEntity, lua: LuaThread) {
	lua.pushTable()
	lua.dup()
	lua.storeGlobal("monster")

	lua.pushBinding(self, "type", ::type)
	lua.pushBinding(self, "seed", ::seed)
	lua.pushBinding(self, "seedNumber", ::seedNumber)
	lua.pushBinding(self, "uniqueParameters", ::uniqueParameters)
	lua.pushBinding(self, "level", ::level)
	lua.pushBinding(self, "setDamageOnTouch", ::setDamageOnTouch)
	lua.pushBinding(self, "setDamageSources", ::setDamageSources)
	lua.pushBinding(self, "setDamageParts", ::setDamageParts)
	lua.pushBinding(self, "setAggressive", ::setAggressive)
	lua.pushBinding(self, "setActiveSkillName", ::setActiveSkillName)
	lua.pushBinding(self, "setDropPool", ::setDropPool)
	lua.pushBinding(self, "toAbsolutePosition", ::toAbsolutePosition)
	lua.pushBinding(self, "mouthPosition", ::mouthPosition)
	lua.pushBinding(self, "flyTo", ::flyTo)
	lua.pushBinding(self, "setDeathParticleBurst", ::setDeathParticleBurst)
	lua.pushBinding(self, "setDeathSound", ::setDeathSound)
	lua.pushBinding(self, "setPhysicsForces", ::setPhysicsForces)
	lua.pushBinding(self, "setName", ::setName)
	lua.pushBinding(self, "setDisplayNametag", ::setDisplayNametag)
	lua.pushBinding(self, false, "say", ::say)
	lua.pushBinding(self, true, "sayPortrait", ::say)
	lua.pushBinding(self, "setDamageTeam", ::setDamageTeam)
	lua.pushBinding(self, "setUniqueId", ::setUniqueId)
	lua.pushBinding(self, "setDamageBar", ::setDamageBar)
	lua.pushBinding(self, "setInteractive", ::setInteractive)
	lua.pushBinding(self, "setAnimationParameter", ::setAnimationParameter)

	lua.pop()
}
