package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonNull
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.EntityDamageTeam
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.util.SBPattern
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.util.valueOfOrNull
import ru.dbotthepony.kstarbound.world.entities.AnchorNetworkState
import ru.dbotthepony.kstarbound.world.entities.HumanoidActorEntity
import ru.dbotthepony.kstarbound.world.entities.LoungeAnchorState
import ru.dbotthepony.kstarbound.world.entities.NPCEntity
import ru.dbotthepony.kstarbound.world.entities.api.LoungeableEntity
import kotlin.collections.contains
import kotlin.collections.indices
import kotlin.collections.isNotEmpty
import kotlin.collections.set
import kotlin.collections.withIndex

private fun toAbsolutePosition(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.toAbsolutePosition(args.nextVector2d()))
	return 1
}

private fun species(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.species.key)
	return 1
}

private fun gender(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.humanoidIdentity.gender.jsonName)
	return 1
}

private fun humanoidIdentity(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(Starbound.gson.toJsonTree(self.variant.humanoidIdentity))
	return 1
}

private fun npcType(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.typeName)
	return 1
}

private fun seed(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.seed)
	return 1
}

private fun level(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.variant.level)
	return 1
}

private fun dropPools(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.pushTable(self.dropPools.size)

	for ((i, entry) in self.dropPools.withIndex()) {
		args.lua.setTableValue(i + 1L, entry.key.left())
	}

	return 1
}

private fun setDropPools(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.dropPools.clear()

	if (args.hasNext) {
		val values = args.readTableValues { Registries.treasurePools.ref(getString(it) ?: throw IllegalArgumentException("Drop pools table contains non-string values")) }
		self.dropPools.addAll(values)
	}

	return 0
}

private fun energy(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.energy)
	return 1
}

private fun maxEnergy(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.maxEnergy)
	return 1
}

private fun say(self: NPCEntity, hasPortrait: Boolean, args: LuaThread.ArgStack): Int {
	val line = args.nextString()
	val portrait = if (hasPortrait) args.nextString() else ""
	val tags = Object2ObjectArrayMap<String, String>()
	val tagsType = args.peek()

	if (tagsType == LuaType.TABLE) {
		val pairs = args.readTable(
			keyVisitor = { getString(it) ?: throw IllegalArgumentException("Tag replacement table contains non-strings as keys") },
			valueVisitor = { getString(it) ?: throw IllegalArgumentException("Tag replacement table contains non-strings as keys") },
		)

		for ((k, v) in pairs) {
			tags[k] = v
		}
	} else if (!tagsType.isNothing) {
		throw IllegalArgumentException("bad argument #2 to object.say: table expected, got $tagsType")
	} else {
		args.skip()
	}

	val sayConfig = args.nextOptionalJson() ?: JsonNull.INSTANCE

	if (line.isBlank()) {
		args.lua.push(false)
	} else if (tags.isEmpty()) {
		self.addChatMessage(line, sayConfig, portrait = portrait)
		args.lua.push(true)
	} else {
		self.addChatMessage(SBPattern.of(line).resolveOrSkip(tags::get), sayConfig, portrait = portrait)
		args.lua.push(true)
	}

	return 1
}

private fun emote(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.addEmote(args.nextString())
	return 0
}

private fun dance(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.setDance(args.nextString())
	return 0
}

private fun setInteractive(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.isInteractive = args.nextBoolean()
	return 0
}

private fun setLounging(self: NPCEntity, args: LuaThread.ArgStack): Int {
	val loungeable = args.nextInt()
	val anchorIndex = args.nextOptionalInt() ?: 0

	val entity = self.world.entities[loungeable] as? LoungeableEntity

	if (entity == null || anchorIndex !in entity.anchors.indices || entity.entitiesLoungingIn(anchorIndex).isNotEmpty()) {
		args.lua.push(false)
	} else {
		self.movement.anchorNetworkState = AnchorNetworkState(loungeable, anchorIndex)
		args.lua.push(true)
	}

	return 1
}

private fun resetLounging(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.movement.anchorNetworkState = null
	return 0
}

private fun isLounging(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.movement.anchorState is LoungeAnchorState)
	return 1
}

private fun loungingIn(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.movement.anchorNetworkState?.entityID?.toLong())
	return 1
}

private fun setItemSlot(self: NPCEntity, args: LuaThread.ArgStack): Int {
	val slot = args.nextString()
	val descriptor = ItemDescriptor(args)
	args.lua.push(self.setItem(slot, descriptor))
	return 1
}

private fun getItemSlot(self: NPCEntity, args: LuaThread.ArgStack): Int {
	val slot = args.nextString()
	val slotType = HumanoidActorEntity.ItemSlot.entries.valueOfOrNull(slot.lowercase())

	if (slotType == null) {
		if (slot in self.variant.items) {
			self.variant.items[slot]!!.store(args.lua)
		} else {
			return 0
		}
	} else {
		self.getItem(slotType).createDescriptor().store(args.lua)
	}

	return 1
}

private fun disableWornArmor(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.disableWornArmor = args.nextBoolean()
	return 0
}

private fun getDisableWornArmor(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.disableWornArmor)
	return 1
}

private fun setShifting(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.isShifting = args.nextBoolean()
	return 0
}

private fun isShifting(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.isShifting)
	return 1
}

private fun getDamageOnTouch(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.damageOnTouch)
	return 1
}

private fun setDamageOnTouch(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.damageOnTouch = args.nextBoolean()
	return 0
}

private fun aimPosition(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.aimPosition)
	return 1
}

private fun setAimPosition(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.aimPosition = self.world.geometry.diff(args.nextVector2d(), self.position)
	return 0
}

private fun getDeathParticleBurst(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.deathParticleBurst)
	return 1
}

private fun setDeathParticleBurst(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.deathParticleBurst = args.nextOptionalString()?.sbIntern()
	return 0
}

private fun getStatusText(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.statusText)
	return 1
}

private fun setStatusText(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.statusText = args.nextOptionalString()?.sbIntern()
	return 0
}

private fun getDisplayNametag(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.displayNametag)
	return 1
}

private fun setDisplayNametag(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.displayNametag = args.nextBoolean()
	return 0
}

private fun getKeepAlive(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.keepAlive)
	return 1
}

private fun setKeepAlive(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.keepAlive = args.nextBoolean()
	return 0
}

private fun getAggressive(self: NPCEntity, args: LuaThread.ArgStack): Int {
	args.lua.push(self.isAggressive)
	return 1
}

private fun setAggressive(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.isAggressive = args.nextBoolean()
	return 0
}

private fun setPersistent(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.isPersistent = args.nextBoolean()
	return 0
}

private fun setDamageTeam(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.team.accept(Starbound.gson.fromJsonFast(args.nextJson(), EntityDamageTeam::class.java))
	return 0
}

private fun setUniqueId(self: NPCEntity, args: LuaThread.ArgStack): Int {
	self.uniqueID.accept(args.nextOptionalString()?.sbIntern())
	return 0
}

fun provideNPCBindings(self: NPCEntity, lua: LuaThread) {
	lua.pushTable()
	lua.dup()
	lua.storeGlobal("npc")

	lua.pushBinding(self, "toAbsolutePosition", ::toAbsolutePosition)
	lua.pushBinding(self, "species", ::species)
	lua.pushBinding(self, "gender", ::gender)
	lua.pushBinding(self, "humanoidIdentity", ::humanoidIdentity)
	lua.pushBinding(self, "npcType", ::npcType)
	lua.pushBinding(self, "seed", ::seed)
	lua.pushBinding(self, "level", ::level)
	lua.pushBinding(self, "dropPools", ::dropPools)
	lua.pushBinding(self, "setDropPools", ::setDropPools)
	lua.pushBinding(self, "energy", ::energy)
	lua.pushBinding(self, "maxEnergy", ::maxEnergy)
	lua.pushBinding(self, false, "say", ::say)
	lua.pushBinding(self, true, "sayPortrait", ::say)
	lua.pushBinding(self, "emote", ::emote)
	lua.pushBinding(self, "dance", ::dance)
	lua.pushBinding(self, "setInteractive", ::setInteractive)
	lua.pushBinding(self, "setLounging", ::setLounging)
	lua.pushBinding(self, "resetLounging", ::resetLounging)
	lua.pushBinding(self, "isLounging", ::isLounging)
	lua.pushBinding(self, "loungingIn", ::loungingIn)
	lua.pushBinding(self, NPCEntity::offeredQuests, "setOfferedQuests", ::setOfferedQuests)
	lua.pushBinding(self, NPCEntity::turnInQuests, "setTurnInQuests", ::setTurnInQuests)
	lua.pushBinding(self, "setItemSlot", ::setItemSlot)
	lua.pushBinding(self, "getItemSlot", ::getItemSlot)
	lua.pushBinding(self, "disableWornArmor", ::disableWornArmor)
	lua.pushBinding(self, "getDisableWornArmor", ::getDisableWornArmor)
	lua.pushBinding(self, "setShifting", ::setShifting)
	lua.pushBinding(self, "isShifting", ::isShifting)
	lua.pushBinding(self, "setDamageOnTouch", ::setDamageOnTouch)
	lua.pushBinding(self, "getDamageOnTouch", ::getDamageOnTouch)
	lua.pushBinding(self, "aimPosition", ::aimPosition)
	lua.pushBinding(self, "setAimPosition", ::setAimPosition)
	lua.pushBinding(self, "getDeathParticleBurst", ::getDeathParticleBurst)
	lua.pushBinding(self, "setDeathParticleBurst", ::setDeathParticleBurst)
	lua.pushBinding(self, "getStatusText", ::getStatusText)
	lua.pushBinding(self, "setStatusText", ::setStatusText)
	lua.pushBinding(self, "getDisplayNametag", ::getDisplayNametag)
	lua.pushBinding(self, "setDisplayNametag", ::setDisplayNametag)
	lua.pushBinding(self, "getKeepAlive", ::getKeepAlive)
	lua.pushBinding(self, "setKeepAlive", ::setKeepAlive)
	lua.pushBinding(self, "getAggressive", ::getAggressive)
	lua.pushBinding(self, "setAggressive", ::setAggressive)
	lua.pushBinding(self, "setPersistent", ::setPersistent)
	lua.pushBinding(self, "setDamageTeam", ::setDamageTeam)
	lua.pushBinding(self, "setUniqueId", ::setUniqueId)

	lua.pushBinding(self, "beginPrimaryFire", NPCEntity::beginPrimaryFire)
	lua.pushBinding(self, "beginAltFire", NPCEntity::beginSecondaryFire)
	lua.pushBinding(self, "beginSecondaryFire", NPCEntity::beginSecondaryFire)
	lua.pushBinding(self, "endPrimaryFire", NPCEntity::endPrimaryFire)
	lua.pushBinding(self, "endAltFire", NPCEntity::endSecondaryFire)
	lua.pushBinding(self, "endSecondaryFire", NPCEntity::endSecondaryFire)

	lua.pop()
}
