package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonNull
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap
import org.apache.logging.log4j.LogManager
import org.classdump.luna.LuaRuntimeException
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.VersionRegistry
import ru.dbotthepony.kstarbound.defs.image.Image
import ru.dbotthepony.kstarbound.defs.image.SpriteReference
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.item.ItemRegistry
import ru.dbotthepony.kstarbound.item.RecipeRegistry
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.nextVector2i
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.util.random.random
import kotlin.collections.set

private val LOGGER = LogManager.getLogger()

private fun <T : Any> lookup(registry: Registry<T>, args: LuaThread.ArgStack): Registry.Entry<T>? {
	return when (val type = args.peek()) {
		LuaType.NUMBER -> registry[args.nextInt()]
		LuaType.STRING -> registry[args.nextString()]
		LuaType.NONE, LuaType.NIL -> null
		else -> throw IllegalArgumentException("Invalid registry key type: $type")
	}
}

private fun <T : Any> lookupStrict(registry: Registry<T>, args: LuaThread.ArgStack): Registry.Entry<T> {
	return when (val type = args.peek()) {
		LuaType.NUMBER -> {
			val key = args.nextInt()
			registry[key] ?: throw LuaRuntimeException("No such ${registry.name}: $key")
		}

		LuaType.STRING -> {
			val key = args.nextString()
			registry[key] ?: throw LuaRuntimeException("No such ${registry.name}: $key")
		}

		else -> throw IllegalArgumentException("Invalid registry key type: $type")
	}
}

private fun registryDef(registry: Registry<*>): LuaThread.Fn {
	return LuaThread.Fn { args ->
		val name = args.nextString()
		val value = registry[name] ?: throw LuaRuntimeException("No such ${registry.name}: $name")
		args.lua.push(value.json)
		return@Fn 1
	}
}

private fun registryDef2(registry: Registry<*>): LuaThread.Fn {
	return LuaThread.Fn { args ->
		val def = lookup(registry, args)

		if (def != null) {
			args.lua.pushTable(hashSize = 2)
			args.lua.setTableValue("path", def.file?.computeFullPath())
			args.lua.setTableValue("config", def.json)

			return@Fn 1
		} else {
			return@Fn 0
		}
	}
}

private fun registryDefExists(registry: Registry<*>): LuaThread.Fn {
	return LuaThread.Fn { args ->
		args.lua.push(args.nextString() in registry)
		return@Fn 1
	}
}

private fun recipesForItem(args: LuaThread.ArgStack): Int {
	val list = RecipeRegistry.output2recipes[args.nextString()]

	if (list == null) {
		args.lua.pushTable()
	} else {
		args.lua.pushTable(list.size)

		for ((i, v) in list.withIndex()) {
			args.lua.setTableValue(i + 1, v.json)
		}
	}

	return 1
}

private fun getMatchingTenants(args: LuaThread.ArgStack): Int {
	val tags = args.lua.readTable(args.position++, { getString() ?: throw IllegalArgumentException("Table contains non-string keys") }, { getLong()?.toInt() ?: throw IllegalArgumentException("Table contains non-integer values") })
		?: throw IllegalArgumentException("bad argument #1 to getMatchingTenants: table expected, got ${args.peek(1)}")

	val actualTags = Object2IntOpenHashMap<String>()

	for ((k, v) in tags) {
		actualTags[k] = v
	}

	val result = Registries.tenants.keys.values
		.stream()
		.filter { it.value.test(actualTags) }
		.sorted { a, b -> b.value.compareTo(a.value) }
		.toList()

	args.lua.pushTable(result.size)

	for ((k, v) in result.withIndex()) {
		args.lua.setTableValue(k + 1, v.key)
	}

	return 1
}

private fun liquidStatusEffects(args: LuaThread.ArgStack): Int {
	val liquid = lookup(Registries.liquid, args)

	if (liquid != null) {
		args.lua.pushTable(liquid.value.statusEffects.size)

		for ((i, effect) in liquid.value.statusEffects.withIndex()) {
			effect.key.map({ args.lua.setTableValue(i + 1L, it) }, { args.lua.setTableValue(i + 1L, it) })
		}
	} else {
		args.lua.pushTable()
	}

	return 1
}

private fun materialMiningSound(args: LuaThread.ArgStack): Int {
	val tile = lookup(Registries.tiles, args)
	val mod = lookup(Registries.tiles, args)

	if (mod != null && mod.value.miningSounds.map({ it.isNotEmpty() }, { true })) {
		args.lua.push(mod.value.miningSounds.map({ it.random(args.lua.random) }, { it }))
	} else if (tile != null && tile.value.miningSounds.map({ it.isNotEmpty() }, { true })) {
		args.lua.push(tile.value.miningSounds.map({ it.random(args.lua.random) }, { it }))
	} else {
		// original engine parity
		args.lua.push("")
	}

	return 1
}

private fun materialFootstepSound(args: LuaThread.ArgStack): Int {
	val tile = lookup(Registries.tiles, args)
	val mod = lookup(Registries.tiles, args)

	if (mod != null && mod.value.footstepSound.map({ it.isNotEmpty() }, { true })) {
		args.lua.push(mod.value.footstepSound.map({ it.random(args.lua.random) }, { it }))
	} else if (tile != null && tile.value.footstepSound.map({ it.isNotEmpty() }, { true })) {
		args.lua.push(tile.value.footstepSound.map({ it.random(args.lua.random) }, { it }))
	} else {
		args.lua.push(Globals.client.defaultFootstepSound.map({ it }, { it.random(args.lua.random) }))
	}

	return 1
}

private fun materialHealth(args: LuaThread.ArgStack): Int {
	args.lua.push(lookupStrict(Registries.tiles, args).value.actualDamageTable.totalHealth)
	return 1
}

private fun liquidName(args: LuaThread.ArgStack): Int {
	args.lua.push(lookupStrict(Registries.liquid, args).key)
	return 1
}

private fun liquidId(args: LuaThread.ArgStack): Int {
	args.lua.push(lookupStrict(Registries.liquid, args).id?.toLong())
	return 1
}

private fun techType(args: LuaThread.ArgStack): Int {
	args.lua.push(lookupStrict(Registries.techs, args).value.type)
	return 1
}

private fun techConfig(args: LuaThread.ArgStack): Int {
	args.lua.push(lookupStrict(Registries.techs, args).json)
	return 1
}

private fun createBiome(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val seed = args.nextLong()
	val verticalMidPoint = args.nextInt()
	val threatLevel = args.nextDouble()

	try {
		val biome = Registries.biomes
			.getOrThrow(name)
			.value
			.create(random(seed), verticalMidPoint, threatLevel)

		args.lua.push(Starbound.gson.toJsonTree(biome))
		return 1
	} catch (err: Throwable) {
		LOGGER.error("Exception while creating biome for Lua script of name $name with seed $seed, with verticalMidPoint $verticalMidPoint, with threatLevel $threatLevel", err)
		return 0
	}
}

private fun treeStemDirectory(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	args.lua.push(Registries.treeStemVariants[name]?.file?.computeDirectory(true) ?: "/")
	return 1
}

private fun treeFoliageDirectory(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	args.lua.push(Registries.treeFoliageVariants[name]?.file?.computeDirectory(true) ?: "/")
	return 1
}

private fun itemConfig(args: LuaThread.ArgStack): Int {
	val desc = ItemDescriptor(args)
	val level = args.nextOptionalDouble()
	val seed = args.nextOptionalLong()

	if (desc.name !in ItemRegistry) {
		return 0
	} else {
		val (config, parameters) = desc.buildConfig(level, seed)

		args.lua.pushTable(hashSize = 3)
		args.lua.setTableValue("directory", ItemRegistry[desc.name].directory)
		args.lua.setTableValue("config", config)
		args.lua.setTableValue("parameters", parameters)

		return 1
	}
}

// why
private fun createItem(args: LuaThread.ArgStack): Int {
	val desc = ItemDescriptor(args)
	val level = args.nextOptionalDouble()
	val seed = args.nextOptionalLong()

	if (desc.name !in ItemRegistry) {
		throw LuaRuntimeException("No such item ${desc.name}")
	} else {
		val (_, params) = desc.buildConfig(level, seed)

		if (desc.store(args.lua, pushParameters = false)) {
			args.lua.setTableValue("parameters", params)
		}

		return 1
	}
}

private fun itemType(args: LuaThread.ArgStack): Int {
	args.lua.push(ItemRegistry.getOrThrow(args.nextString()).type.jsonName)
	return 1
}

private fun itemTags(args: LuaThread.ArgStack): Int {
	val tags = ItemRegistry.getOrThrow(args.nextString()).itemTags
	args.lua.pushTable(tags.size)

	for ((i, tag) in tags.withIndex()) {
		args.lua.setTableValue(i + 1L, tag)
	}

	return 1
}

private fun itemHasTag(args: LuaThread.ArgStack): Int {
	val item = args.nextString()
	val tag = args.nextString()

	args.lua.push(tag in ItemRegistry.getOrThrow(item).itemTags)
	return 1
}

private fun monsterSkillParameter(args: LuaThread.ArgStack): Int {
	val skillName = args.nextString()
	val configParameterName = args.nextString()
	val skill = Registries.monsterSkills[skillName]

	if (skill != null) {
		args.lua.push(skill.value.config[configParameterName] ?: JsonNull.INSTANCE)
		return 1
	} else {
		return 0
	}
}

private fun monsterParameters(args: LuaThread.ArgStack): Int {
	val monsterType = args.nextString()
	val seed = args.nextOptionalLong()
	args.lua.push(Registries.monsterTypes.getOrThrow(monsterType).value.create(seed ?: 0L, JsonObject()).parameters)
	return 1
}

private fun monsterMovementSettings(args: LuaThread.ArgStack): Int {
	val monsterType = args.nextString()
	val seed = args.nextOptionalLong()
	args.lua.push(Registries.monsterTypes.getOrThrow(monsterType).value.create(seed ?: 0L, JsonObject()).parameters["movementSettings"] ?: JsonObject())
	return 1
}

private fun elementalResistance(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val type = Registries.damageKinds.getOrThrow(name).value.elementalType
	val elemental = Globals.elementalTypes[type] ?: throw NoSuchElementException("Damage kind $name has specified $type as its elemental damage type, but it is missing from /damage/elementaltypes.config")
	args.lua.push(elemental.resistanceStat)
	return 1
}

private fun dungeonMetadata(args: LuaThread.ArgStack): Int {
	args.lua.push(Registries.dungeons.getOrThrow(args.nextString()).jsonObject["metadata"])
	return 1
}

private val hasTech = registryDefExists(Registries.techs)

private fun assetJson(args: LuaThread.ArgStack): Int {
	args.lua.push(Starbound.loadJsonAsset(args.nextString()).get())
	return 1
}

private fun makeCurrentVersionedJson(args: LuaThread.ArgStack): Int {
	args.lua.push(VersionRegistry.make(args.nextString(), args.nextJson()).toJson())
	return 1
}

private fun loadVersionedJson(args: LuaThread.ArgStack): Int {
	args.lua.push(VersionRegistry.load(args.nextString(), args.nextJson()))
	return 1
}

private fun evalFunction(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val value = args.nextDouble()
	val fn = Registries.jsonFunctions[name] ?: throw LuaRuntimeException("No such function $name")
	args.lua.push(fn.value.evaluate(value))
	return 1
}

private fun evalFunction2(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val value = args.nextDouble()
	val value2 = args.nextDouble()
	val fn = Registries.json2Functions[name] ?: throw LuaRuntimeException("No such function $name")
	args.lua.push(fn.value.evaluate(value, value2))
	return 1
}

private fun imageSize(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val ref = SpriteReference.create(name)
	val sprite = ref.sprite ?: throw LuaRuntimeException("No such image or sprite $ref")
	args.lua.pushTable(2)
	args.lua.setTableValue(1, sprite.width)
	args.lua.setTableValue(2, sprite.height)
	return 1
}

private fun imageSpaces(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val image = Image.get(name) ?: throw LuaRuntimeException("No such image $name")

	val pixelOffset = args.nextVector2i()
	val fillFactor = args.nextDouble()
	val flip = args.nextOptionalBoolean() ?: false

	val values = image.worldSpaces(pixelOffset, fillFactor, flip)
	args.lua.pushTable(values.size, 0)

	for ((i, value) in values.withIndex()) {
		args.lua.push(i + 1L)
		args.lua.push(value)
		args.lua.setTableValue()
	}

	return 1
}

private fun nonEmptyRegion(args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val image = Image.get(name) ?: throw LuaRuntimeException("No such image $name")
	args.lua.push(image.nonEmptyRegion)
	return 1
}

private fun generateName(args: LuaThread.ArgStack): Int {
	val asset = args.nextString()
	val seed = args.nextOptionalLong()
	args.lua.push(Starbound.generateName(asset, if (seed == null) args.lua.random else random(seed.toLong())))
	return 1
}

private fun createTreasure(args: LuaThread.ArgStack): Int {
	val pool = args.nextString()
	val level = args.nextDouble()
	val seed = args.nextOptionalLong()

	val get = Registries.treasurePools.getOrThrow(pool)
	val random = if (seed == null) args.lua.random else random(seed.toLong())

	args.lua.pushTable()
	var i = 1L

	for (loot in get.value.evaluate(random, level)) {
		if (loot.isNotEmpty) {
			args.lua.push(i++)
			loot.createDescriptor().store(args.lua)
			args.lua.setTableValue()
		}
	}

	return 1
}

private val projectileConfig = registryDef(Registries.projectiles)
private val tenantConfig = registryDef(Registries.tenants)
private val questConfig = registryDef(Registries.questTemplates)
private val isTreasurePool = registryDefExists(Registries.treasurePools)

private val materialConfig = registryDef2(Registries.tiles)
private val modConfig = registryDef2(Registries.tileModifiers)

fun provideRootBindings(lua: LuaThread) {
	lua.pushTable()
	lua.dup()
	lua.storeGlobal("root")

	lua.setTableValue("assetJson", ::assetJson)

	lua.setTableValue("makeCurrentVersionedJson", ::makeCurrentVersionedJson)
	lua.setTableValue("loadVersionedJson", ::loadVersionedJson)

	lua.setTableValue("evalFunction", ::evalFunction)
	lua.setTableValue("evalFunction2", ::evalFunction2)

	lua.setTableValue("imageSize", ::imageSize)
	lua.setTableValue("imageSpaces", ::imageSpaces)
	lua.setTableValue("nonEmptyRegion", ::nonEmptyRegion)

	lua.setTableValueToStub("npcConfig")
	lua.setTableValueToStub("npcVariant")
	lua.setTableValueToStub("projectileGravityMultiplier")
	lua.setTableValueToStub("npcPortrait")
	lua.setTableValueToStub("monsterPortrait")

	lua.setTableValue("recipesForItem", ::recipesForItem)
	lua.setTableValue("itemType", ::itemType)
	lua.setTableValue("itemTags", ::itemTags)
	lua.setTableValue("itemHasTag", ::itemHasTag)
	lua.setTableValue("itemConfig", ::itemConfig)
	lua.setTableValue("createItem", ::createItem)

	lua.setTableValue("projectileConfig", projectileConfig)
	lua.setTableValue("tenantConfig", tenantConfig)

	lua.setTableValue("getMatchingTenants", ::getMatchingTenants)
	lua.setTableValue("liquidStatusEffects", ::liquidStatusEffects)

	lua.setTableValue("generateName", ::generateName)
	lua.setTableValue("questConfig", questConfig)

	lua.setTableValue("isTreasurePool", isTreasurePool)
	lua.setTableValue("createTreasure", ::createTreasure)

	lua.setTableValue("materialMiningSound", ::materialMiningSound)
	lua.setTableValue("materialFootstepSound", ::materialFootstepSound)
	lua.setTableValue("materialHealth", ::materialHealth)

	lua.setTableValue("materialConfig", materialConfig)
	lua.setTableValue("modConfig", modConfig)

	lua.setTableValue("liquidName", ::liquidName)
	lua.setTableValue("liquidId", ::liquidId)

	lua.setTableValue("createBiome", ::createBiome)
	lua.setTableValue("monsterSkillParameter", ::monsterSkillParameter)
	lua.setTableValue("monsterParameters", ::monsterParameters)
	lua.setTableValue("monsterMovementSettings", ::monsterMovementSettings)

	lua.setTableValue("hasTech", hasTech)
	lua.setTableValue("techType", ::techType)
	lua.setTableValue("techType", ::techConfig)

	lua.setTableValue("treeStemDirectory", ::treeStemDirectory)
	lua.setTableValue("treeFoliageDirectory", ::treeFoliageDirectory)

	lua.setTableValueToStub("collection")
	lua.setTableValueToStub("collectables")

	lua.setTableValue("elementalResistance", ::elementalResistance)
	lua.setTableValue("dungeonMetadata", ::dungeonMetadata)

	lua.pop()
}
