package ru.dbotthepony.kstarbound.lua.bindings

import org.apache.logging.log4j.LogManager
import org.classdump.luna.LuaRuntimeException
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.defs.tile.NO_DUNGEON_ID
import ru.dbotthepony.kstarbound.defs.tile.isNotEmptyLiquid
import ru.dbotthepony.kstarbound.defs.world.BiomeDefinition
import ru.dbotthepony.kstarbound.defs.world.BiomePlaceables
import ru.dbotthepony.kstarbound.defs.world.BiomePlaceablesDefinition
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.LuaType
import ru.dbotthepony.kstarbound.lua.nextAABB
import ru.dbotthepony.kstarbound.lua.nextAABBi
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.nextVector2i
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.lua.userdata.LuaFuture
import ru.dbotthepony.kstarbound.lua.userdata.push
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.util.random.staticRandom64
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.world.ChunkState
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.ItemDropEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import java.util.concurrent.CompletableFuture
import java.util.concurrent.TimeUnit

private val LOGGER = LogManager.getLogger()

private fun placeDungeonImpl(self: ServerWorld, force: Boolean, async: Boolean, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val pos = args.nextVector2i()
	val dungeonID = args.nextOptionalLong()?.toInt()
	val seed = args.nextOptionalLong()
	val actualSeed = seed ?: staticRandom64(pos.x, pos.y, "DungeonPlacement")
	val dungeonDef = Registries.dungeons[name]

	if (dungeonDef == null) {
		LOGGER.error("Lua script tried to spawn dungeon $name, however, no such dungeon exist")

		if (async) {
			args.lua.push(LuaFuture(
				future = CompletableFuture.failedFuture(NoSuchElementException("No such dungeon: $name")),
				isLocal = false,
				handler = { 0 }
			))
		} else {
			args.lua.push(false)
		}
	} else {
		val future = dungeonDef.value.generate(
			self,
			random(actualSeed),
			pos.x, pos.y,
			markSurfaceAndTerrain = false,
			forcePlacement = force,
			dungeonID = dungeonID?.toInt() ?: NO_DUNGEON_ID,
			// Setting scope to eventloop will make dungeon generate synchronous if it is possible
			// It is still possible, though, for dungeon to not generate synchronously if it need to load chunks
			// which are currently not loaded
			scope = if (async) Starbound.GLOBAL_SCOPE else self.eventLoop.scope).thenApply { it.hasGenerated }

		if (async) {
			args.lua.push(LuaFuture(
				future = future,
				isLocal = false,
				handler = { push(it); 1 }
			))
		} else {
			args.lua.push(future.getNow(true))
		}
	}

	return 1
}

private fun breakObject(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? WorldObject
	val smash = args.nextOptionalBoolean()

	if (entity == null) {
		args.lua.push(false)
	} else if (entity.isRemote) {
		// we can't break objects now owned by us
		args.lua.push(false)
	} else {
		if (smash == true)
			entity.health = 0.0

		entity.remove(AbstractEntity.RemovalReason.DYING)
		args.lua.push(true)
	}

	return 1
}

private fun isVisibleToPlayer(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val region = args.nextAABB()
	args.lua.push(self.clients.any { it.isTracking(region) })
	return 1
}

private fun loadRegion(self: ServerWorld, args: LuaThread.ArgStack): Int {
	// FIXME: in original engine this supposedly blocks everything else from happening until region is loaded
	//  And we can't block event loop, because this will make region be unable to be loaded in first place
	val region = args.nextAABB()

	// keep in ram for at most 2400 ticks
	val tickets = self.temporaryChunkTicket(region, 2400).get()
	val future = CompletableFuture.allOf(*tickets.map { it.chunk }.toTypedArray())

	future.thenApply {
		self.eventLoop.schedule(Runnable {
			tickets.forEach { it.cancel() }
		}, 4L, TimeUnit.SECONDS)
	}.exceptionally {
		self.eventLoop.schedule(Runnable {
			tickets.forEach { it.cancel() }
		}, 4L, TimeUnit.SECONDS)
	}

	args.lua.push(LuaFuture(
		future = future,
		isLocal = false,
		handler = { 0 }
	))

	return 1
}

private fun regionActive(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val chunks = self.geometry.region2Chunks(args.nextAABB())
	// TODO: i have no idea what mods expect this to return, so i'll have to guess
	args.lua.push(chunks.all { self.chunkMap[it]?.state == ChunkState.FULL })
	return 1
}

private fun setTileProtection(self: ServerWorld, args: LuaThread.ArgStack): Int {
	self.switchDungeonIDProtection(args.nextInt(), args.nextBoolean())
	return 0
}

private fun isPlayerModified(self: ServerWorld, args: LuaThread.ArgStack): Int {
	args.lua.push(self.isPlayerModified(args.nextAABBi()))
	return 1
}

private fun forceDestroyLiquid(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val position = args.nextVector2i()
	val cell = self.getCell(position).mutable()

	if (cell.liquid.state.isNotEmptyLiquid && cell.liquid.level > 0f) {
		args.lua.push(cell.liquid.level.toDouble())

		cell.liquid.reset()
		self.setCell(position, cell.immutable())
		return 1
	}

	return 0
}

private fun loadUniqueEntityAsync(self: ServerWorld, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = self.loadUniqueEntity(args.nextString()).thenApply { KOptional.ofNullable(it) },
			isLocal = false,
			handler = {
				push(it.orNull()?.entityID?.toLong())
				1
			}
		)
	)

	return 1
}

private fun setUniqueId(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val name = args.nextString()
	val entity = self.entities[id] ?: throw IllegalStateException("No such entity with ID $id (tried to set unique id to $name)")

	if (entity.isRemote)
		throw IllegalStateException("Entity $entity is not owned by this side")

	if (
		entity.type == EntityType.NPC ||
		entity.type == EntityType.STAGEHAND ||
		entity.type == EntityType.MONSTER ||
		entity.type == EntityType.OBJECT
	) {
		entity.uniqueID.accept(name.sbIntern())
	} else {
		throw IllegalStateException("Entity type is restricted from having unique ID: $entity (tried to set unique id to $name)")
	}

	return 0
}

private fun takeItemDrop(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val entity = self.entities[id] as? ItemDropEntity ?: return 0
	val takenBy = args.nextOptionalLong()?.toInt()

	if (entity.isLocal && entity.canTake) {
		entity.take(takenBy ?: 0).createDescriptor().store(args.lua)
		return 1
	}

	return 0
}

private fun setPlayerStart(self: ServerWorld, args: LuaThread.ArgStack): Int {
	self.setPlayerSpawn(args.nextVector2d(), args.nextOptionalBoolean() == true)
	return 0
}

private fun players(self: ServerWorld, args: LuaThread.ArgStack): Int {
	args.lua.pushTable(self.clients.size)
	var i = 1L

	for (p in self.clients) {
		val entityID = p.client.playerEntity?.entityID

		if (entityID != null)
			args.lua.setTableValue(i++, entityID)
	}

	return 1
}

private fun setSkyTime(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val time = args.nextDouble()

	if (time < 0.0)
		throw LuaRuntimeException("Negative time? $time")

	self.sky.time = time
	return 0
}

private fun setUniverseFlag(self: ServerWorld, args: LuaThread.ArgStack): Int {
	args.lua.push(self.server.addUniverseFlag(args.nextString().sbIntern()))
	return 1
}

private fun unsetUniverseFlag(self: ServerWorld, args: LuaThread.ArgStack): Int {
	args.lua.push(self.server.removeUniverseFlag(args.nextString()))
	return 1
}

private fun universeFlags(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val flags = self.server.getUniverseFlags()

	args.lua.pushTable(flags.size)

	for ((i, flag) in flags.withIndex()) {
		args.lua.setTableValue(i + 1L, flag)
	}

	return 1
}

private fun universeFlagSet(self: ServerWorld, args: LuaThread.ArgStack): Int {
	args.lua.push(self.server.hasUniverseFlag(args.nextString()))
	return 1
}

private fun placeDungeon(self: ServerWorld, args: LuaThread.ArgStack): Int {
	return placeDungeonImpl(self, force = true, async = false, args)
}

private fun tryPlaceDungeon(self: ServerWorld, args: LuaThread.ArgStack): Int {
	return placeDungeonImpl(self, force = false, async = false, args)
}

private fun placeDungeonAsync(self: ServerWorld, args: LuaThread.ArgStack): Int {
	return placeDungeonImpl(self, force = true, async = true, args)
}

private fun tryPlaceDungeonAsync(self: ServerWorld, args: LuaThread.ArgStack): Int {
	return placeDungeonImpl(self, force = false, async = true, args)
}

private fun setDungeonGravity(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val id = args.nextInt()
	val peek = args.peek()

	if (peek.isNothing) {
		self.setDungeonGravity(id, null)
	} else if (peek == LuaType.NUMBER) {
		self.setDungeonGravity(id, args.nextDouble())
	} else if (peek == LuaType.TABLE) {
		self.setDungeonGravity(id, args.nextVector2d())
	} else {
		throw LuaRuntimeException("Illegal gravity argument to setDungeonGravity: $peek")
	}

	return 0
}

private fun setDungeonBreathable(self: ServerWorld, args: LuaThread.ArgStack): Int {
	self.setDungeonBreathable(args.nextInt(), args.nextOptionalBoolean())
	return 0
}

private fun setDungeonId(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val region = args.nextAABBi()
	val id = args.nextInt()

	for (x in region.mins.x .. region.maxs.x) {
		for (y in region.mins.y .. region.maxs.y) {
			val cell = self.getCell(x, y)

			if (cell.dungeonId != id) {
				self.setCell(x, y, cell.mutable().also { it.dungeonId = id })
			}
		}
	}

	return 0
}

private fun enqueuePlacement(self: ServerWorld, args: LuaThread.ArgStack): Int {
	val distributions = args.nextJsonArray()
	val dungeonId = args.nextOptionalInt()
	val items = ArrayList<BiomePlaceables.DistributionItem>()

	for (v in distributions) {
		// original engine treats distributions table as if it was originating from biome json files
		val unprepared = Starbound.gson.fromJson(toJsonFromLua(v), BiomePlaceablesDefinition.DistributionItem::class.java)
		val prepared = unprepared.create(BiomeDefinition.CreationParams(hueShift = 0.0, random = args.lua.random))
		items.add(prepared)
	}

	// TODO: Enqueued placements are lost on world shutdown.
	//  Shouldn't we serialize them to persistent storage?
	args.lua.push(
		LuaFuture(
			future = self.enqueuePlacement(items, dungeonId),
			isLocal = false,
			handler = { push(it); 1 }
		)
	)

	return 1
}

private val script by lazy { LuaThread.loadInternalScript("server_world") }

fun provideServerWorldBindings(self: ServerWorld, lua: LuaThread) {
	lua.pushBinding(self, "breakObject", ::breakObject)
	lua.pushBinding(self, "isVisibleToPlayer", ::isVisibleToPlayer)
	lua.pushBinding(self, "loadRegion", ::loadRegion)
	lua.pushBinding(self, "regionActive", ::regionActive)
	lua.pushBinding(self, "setTileProtection", ::setTileProtection)
	lua.pushBinding(self, "isPlayerModified", ::isPlayerModified)
	lua.pushBinding(self, "forceDestroyLiquid", ::forceDestroyLiquid)
	lua.pushBinding(self, "loadUniqueEntityAsync", ::loadUniqueEntityAsync)
	lua.pushBinding(self, "setUniqueId", ::setUniqueId)
	lua.pushBinding(self, "takeItemDrop", ::takeItemDrop)
	lua.pushBinding(self, "setPlayerStart", ::setPlayerStart)
	lua.pushBinding(self, "players", ::players)
	lua.pushBinding(self, "setSkyTime", ::setSkyTime)
	lua.pushBinding(self, "setUniverseFlag", ::setUniverseFlag)
	lua.pushBinding(self, "unsetUniverseFlag", ::unsetUniverseFlag)
	lua.pushBinding(self, "universeFlags", ::universeFlags)
	lua.pushBinding(self, "universeFlagSet", ::universeFlagSet)
	lua.pushBinding(self, "placeDungeon", ::placeDungeon)
	lua.pushBinding(self, "tryPlaceDungeon", ::tryPlaceDungeon)
	lua.pushBinding(self, "placeDungeonAsync", ::placeDungeonAsync)
	lua.pushBinding(self, "tryPlaceDungeonAsync", ::tryPlaceDungeonAsync)
	lua.pushBinding(self, "setDungeonGravity", ::setDungeonGravity)
	lua.pushBinding(self, "setDungeonBreathable", ::setDungeonBreathable)
	lua.pushBinding(self, "setDungeonId", ::setDungeonId)
	lua.pushBinding(self, "enqueuePlacement", ::enqueuePlacement)

	// terraforming
	lua.setTableValueToStub("addBiomeRegion")
	lua.setTableValueToStub("expandBiomeRegion")
	lua.setTableValueToStub("pregenerateAddBiome")
	lua.setTableValueToStub("pregenerateExpandBiome")
	lua.setTableValueToStub("setLayerEnvironmentBiome")
	lua.setTableValueToStub("setPlanetType")

	lua.load(script, "@/internal/server_world.lua")
	lua.call()
}
