package ru.dbotthepony.kstarbound.lua.bindings

import com.google.gson.JsonNull
import com.google.gson.reflect.TypeToken
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.DamageData
import ru.dbotthepony.kstarbound.defs.EphemeralStatusEffect
import ru.dbotthepony.kstarbound.defs.actor.PersistentStatusEffect
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.util.sbIntern
import ru.dbotthepony.kstarbound.world.entities.StatusController

private object PersistentStatusEffectToken : TypeToken<PersistentStatusEffect>()
private object CPersistentStatusEffectToken : TypeToken<ArrayList<PersistentStatusEffect>>()

private fun statusProperty(self: StatusController, args: LuaThread.ArgStack): Int {
	val get = self.getProperty(args.nextString())

	if (get == null) {
		if (args.top == 1) {
			args.lua.push()
		} else if (args.top > 3) {
			args.lua.dup(2)
		}
	} else {
		args.lua.push(get)
	}

	return 1
}

private fun setStatusProperty(self: StatusController, args: LuaThread.ArgStack): Int {
	self.setProperty(args.nextString().sbIntern(), args.nextOptionalJson() ?: JsonNull.INSTANCE)
	return 0
}

private fun stat(self: StatusController, args: LuaThread.ArgStack): Int {
	args.lua.push(self.effectiveStats[args.nextString()]?.effectiveModifiedValue ?: 0.0)
	return 1
}

private fun statPositive(self: StatusController, args: LuaThread.ArgStack): Int {
	args.lua.push(self.statPositive(args.nextString()))
	return 1
}

private fun resourceNames(self: StatusController, args: LuaThread.ArgStack): Int {
	args.lua.pushTable(self.resources.size)

	for ((i, value) in self.resources.keys.withIndex())
		args.lua.setTableValue(i + 1, value)

	return 1
}

private fun isResource(self: StatusController, args: LuaThread.ArgStack): Int {
	args.lua.push(args.nextString() in self.resources)
	return 1
}

private fun resource(self: StatusController, args: LuaThread.ArgStack): Int {
	val resource = self.resources[args.nextString()] ?: return 0
	args.lua.push(resource.value)
	return 1
}

private fun resourcePositive(self: StatusController, args: LuaThread.ArgStack): Int {
	val resource = self.resources[args.nextString()]

	if (resource == null)
		args.lua.push(false)
	else
		args.lua.push(resource.isPositive)

	return 1
}

private fun setResource(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	resource.value = args.nextDouble()
	return 0
}

private fun modifyResource(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	resource.value += args.nextDouble()
	return 0
}

private fun giveResource(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name]
	val amount = args.nextDouble()

	if (resource == null) {
		args.lua.push(0.0)
	} else {
		args.lua.push(resource.give(amount))
	}

	return 1
}

private fun consumeResource(self: StatusController, overdraw: Boolean, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name]
	val amount = args.nextDouble()

	if (resource == null) {
		args.lua.push(false)
	} else {
		args.lua.push(resource.consume(amount, overdraw))
	}

	return 1
}

private fun resourceLocked(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	args.lua.push(resource.isLocked)
	return 1
}

private fun setResourceLocked(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	resource.isLocked = args.nextBoolean()
	return 0
}

private fun resetResource(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	resource.reset()
	return 0
}

private fun resetAllResources(self: StatusController, args: LuaThread.ArgStack): Int {
	self.resetResources()
	return 0
}

private fun resourceMax(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	args.lua.push(resource.maxValue)
	return 1
}

private fun resourcePercentage(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	args.lua.push(resource.percentage)
	return 1
}

private fun setResourcePercentage(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	resource.setAsPercentage(args.nextDouble())
	return 0
}

private fun modifyResourcePercentage(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val resource = self.resources[name] ?: throw NoSuchElementException("No such resource $name")
	resource.modifyPercentage(args.nextDouble())
	return 0
}

private fun getPersistentEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	val effects = self.getPersistentEffects(args.nextString())
	args.lua.pushTable(effects.size)

	for ((i, effect) in effects.withIndex()) {
		args.lua.push(i + 1L)
		effect.map({ args.lua.push(Starbound.gson.toJsonTree(it)) }, { args.lua.push(it) })
		args.lua.setTableValue()
	}

	return 1
}

private fun addPersistentEffect(self: StatusController, args: LuaThread.ArgStack): Int {
	val category = args.nextString().sbIntern()
	val effect = Starbound.gson.fromJsonFast(args.nextJson(), PersistentStatusEffectToken)
	self.addPersistentEffect(category, effect)
	return 0
}

private fun addPersistentEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	val category = args.nextString().sbIntern()
	val effects = Starbound.gson.fromJsonFast(args.nextJson(), CPersistentStatusEffectToken)
	self.addPersistentEffects(category, effects)
	return 0
}

private fun setPersistentEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	val category = args.nextString().sbIntern()
	val effects = Starbound.gson.fromJsonFast(args.nextJson(), CPersistentStatusEffectToken)
	self.setPersistentEffects(category, effects)
	return 0
}

private fun clearPersistentEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	val category = args.nextString()
	self.removePersistentEffects(category)
	return 0
}

private fun clearAllPersistentEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	self.removeAllPersistentEffects()
	return 0
}

private fun addEphemeralEffect(self: StatusController, args: LuaThread.ArgStack): Int {
	val name = args.nextString()
	val duration = args.nextOptionalDouble()
	val sourceEntity = args.nextOptionalInt()

	self.addEphemeralEffect(EphemeralStatusEffect(Registries.statusEffects.ref(name), duration), sourceEntity)
	return 0
}

private fun addEphemeralEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	val effects = args.readTableValues { Starbound.gson.fromJsonFast(getJson() ?: throw IllegalArgumentException("invalid values in effects table"), EphemeralStatusEffect::class.java) }
	val sourceEntity = args.nextOptionalInt()

	for (effect in effects) {
		self.addEphemeralEffect(effect, sourceEntity)
	}

	return 0
}

private fun removeEphemeralEffect(self: StatusController, args: LuaThread.ArgStack): Int {
	self.removeEphemeralEffect(args.nextString())
	return 0
}

private fun removeEphemeralEffects(self: StatusController, args: LuaThread.ArgStack): Int {
	self.removeEphemeralEffects()
	return 0
}

private fun damageTakenSince(self: StatusController, args: LuaThread.ArgStack): Int {
	val since = args.nextOptionalLong() ?: 0L
	val (list, newSince) = self.recentDamageReceived(since)

	args.lua.pushTable(list.size)

	for ((i, v) in list.withIndex()) {
		args.lua.setTableValue(i + 1L, Starbound.gson.toJsonTree(v))
	}

	args.lua.push(newSince)
	return 2
}

private fun inflictedHitsSince(self: StatusController, args: LuaThread.ArgStack): Int {
	val since = args.nextOptionalLong() ?: 0L
	val (list, newSince) = self.recentHitsDealt(since)

	args.lua.pushTable(list.size)

	for ((i, v) in list.withIndex()) {
		val (target, data) = v

		args.lua.push(i + 1L)

		args.lua.push(Starbound.gson.toJsonTree(data))
		args.lua.setTableValue("targetEntityId", target)

		args.lua.setTableValue()
	}

	args.lua.push(newSince)
	return 2
}

private fun inflictedDamageSince(self: StatusController, args: LuaThread.ArgStack): Int {
	val since = args.nextOptionalLong() ?: 0L
	val (list, newSince) = self.recentHitsDealt(since)

	args.lua.pushTable(list.size)

	for ((i, v) in list.withIndex()) {
		args.lua.setTableValue(i + 1L, Starbound.gson.toJsonTree(v))
	}

	args.lua.push(newSince)
	return 2
}

private fun activeUniqueStatusEffectSummary(self: StatusController, args: LuaThread.ArgStack): Int {
	val effects = self.activeUniqueStatusEffectSummary()

	args.lua.pushTable(effects.size)

	for ((i, data) in effects.withIndex()) {
		args.lua.push(i + 1L)
		args.lua.pushTable(2)
		args.lua.setTableValue(1L, data.first.key)
		args.lua.setTableValue(2L, data.second)
		args.lua.setTableValue()
	}

	return 1
}

private fun uniqueStatusEffectActive(self: StatusController, args: LuaThread.ArgStack): Int {
	args.lua.push(self.uniqueStatusEffectActive(args.nextString()))
	return 1
}

private fun primaryDirectives(self: StatusController, args: LuaThread.ArgStack): Int {
	args.lua.push(self.primaryDirectives)
	return 1
}

private fun setPrimaryDirectives(self: StatusController, args: LuaThread.ArgStack): Int {
	self.primaryDirectives = args.nextOptionalString()?.sbIntern() ?: ""
	return 0
}

private fun applySelfDamageRequest(self: StatusController, args: LuaThread.ArgStack): Int {
	self.inflictSelfDamage(Starbound.gson.fromJsonFast(args.nextJson(), DamageData::class.java))
	return 1
}

private fun appliesEnvironmentStatusEffects(self: StatusController, args: LuaThread.ArgStack): Int { args.lua.push(self.appliesEnvironmentStatusEffects); return 1 }
private fun appliesWeatherStatusEffects(self: StatusController, args: LuaThread.ArgStack): Int { args.lua.push(self.appliesWeatherStatusEffects); return 1 }
private fun minimumLiquidStatusEffectPercentage(self: StatusController, args: LuaThread.ArgStack): Int { args.lua.push(self.minimumLiquidStatusEffectPercentage); return 1 }

private fun setAppliesEnvironmentStatusEffects(self: StatusController, args: LuaThread.ArgStack): Int { self.appliesEnvironmentStatusEffects = args.nextBoolean(); return 0 }
private fun setAppliesWeatherStatusEffects(self: StatusController, args: LuaThread.ArgStack): Int { self.appliesWeatherStatusEffects = args.nextBoolean(); return 0 }
private fun setMinimumLiquidStatusEffectPercentage(self: StatusController, args: LuaThread.ArgStack): Int { self.minimumLiquidStatusEffectPercentage = args.nextDouble(); return 0 }

fun provideStatusControllerBindings(self: StatusController, lua: LuaThread) {
	lua.pushTable()
	lua.dup()
	lua.storeGlobal("status")

	lua.pushBinding(self, "statusProperty", ::statusProperty)
	lua.pushBinding(self, "setStatusProperty", ::setStatusProperty)
	lua.pushBinding(self, "stat", ::stat)
	lua.pushBinding(self, "statPositive", ::statPositive)
	lua.pushBinding(self, "resourceNames", ::resourceNames)
	lua.pushBinding(self, "isResource", ::isResource)
	lua.pushBinding(self, "resource", ::resource)
	lua.pushBinding(self, "resourcePositive", ::resourcePositive)
	lua.pushBinding(self, "setResource", ::setResource)
	lua.pushBinding(self, "modifyResource", ::modifyResource)
	lua.pushBinding(self, "giveResource", ::giveResource)
	lua.pushBinding(self, false, "consumeResource", ::consumeResource)
	lua.pushBinding(self, true, "overConsumeResource", ::consumeResource)
	lua.pushBinding(self, "resourceLocked", ::resourceLocked)
	lua.pushBinding(self, "setResourceLocked", ::setResourceLocked)
	lua.pushBinding(self, "resetResource", ::resetResource)
	lua.pushBinding(self, "resetAllResources", ::resetAllResources)
	lua.pushBinding(self, "resourceMax", ::resourceMax)
	lua.pushBinding(self, "resourcePercentage", ::resourcePercentage)
	lua.pushBinding(self, "setResourcePercentage", ::setResourcePercentage)
	lua.pushBinding(self, "modifyResourcePercentage", ::modifyResourcePercentage)
	lua.pushBinding(self, "getPersistentEffects", ::getPersistentEffects)
	lua.pushBinding(self, "addPersistentEffect", ::addPersistentEffect)
	lua.pushBinding(self, "addPersistentEffects", ::addPersistentEffects)
	lua.pushBinding(self, "setPersistentEffects", ::setPersistentEffects)
	lua.pushBinding(self, "clearPersistentEffects", ::clearPersistentEffects)
	lua.pushBinding(self, "clearAllPersistentEffects", ::clearAllPersistentEffects)
	lua.pushBinding(self, "addEphemeralEffect", ::addEphemeralEffect)
	lua.pushBinding(self, "addEphemeralEffects", ::addEphemeralEffects)
	lua.pushBinding(self, "removeEphemeralEffect", ::removeEphemeralEffect)
	lua.pushBinding(self, "removeEphemeralEffects", ::removeEphemeralEffects)
	lua.pushBinding(self, "damageTakenSince", ::damageTakenSince)
	lua.pushBinding(self, "inflictedHitsSince", ::inflictedHitsSince)
	lua.pushBinding(self, "inflictedDamageSince", ::inflictedDamageSince)
	lua.pushBinding(self, "activeUniqueStatusEffectSummary", ::activeUniqueStatusEffectSummary)
	lua.pushBinding(self, "uniqueStatusEffectActive", ::uniqueStatusEffectActive)
	lua.pushBinding(self, "primaryDirectives", ::primaryDirectives)
	lua.pushBinding(self, "setPrimaryDirectives", ::setPrimaryDirectives)
	lua.pushBinding(self, "applySelfDamageRequest", ::applySelfDamageRequest)
	lua.pushBinding(self, "appliesEnvironmentStatusEffects", ::appliesEnvironmentStatusEffects)
	lua.pushBinding(self, "appliesWeatherStatusEffects", ::appliesWeatherStatusEffects)
	lua.pushBinding(self, "minimumLiquidStatusEffectPercentage", ::minimumLiquidStatusEffectPercentage)
	lua.pushBinding(self, "setAppliesEnvironmentStatusEffects", ::setAppliesEnvironmentStatusEffects)
	lua.pushBinding(self, "setAppliesWeatherStatusEffects", ::setAppliesWeatherStatusEffects)
	lua.pushBinding(self, "setMinimumLiquidStatusEffectPercentage", ::setMinimumLiquidStatusEffectPercentage)

	lua.pop()
}
