package ru.dbotthepony.kstarbound.lua.bindings

import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.tile.TileDamageType
import ru.dbotthepony.kstarbound.defs.tile.isEmptyTile
import ru.dbotthepony.kstarbound.defs.tile.isNotEmptyModifier
import ru.dbotthepony.kstarbound.defs.tile.isNullTile
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.getVector2i
import ru.dbotthepony.kstarbound.lua.nextVector2d
import ru.dbotthepony.kstarbound.lua.nextVector2i
import ru.dbotthepony.kstarbound.lua.push
import ru.dbotthepony.kstarbound.lua.userdata.LuaFuture
import ru.dbotthepony.kstarbound.lua.userdata.push
import ru.dbotthepony.kstarbound.math.vector.Vector2i
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.TileModification
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.api.TileColor
import ru.dbotthepony.kstarbound.world.tileAreaBrush
import java.util.concurrent.CompletableFuture

private fun damageTilesImpl(self: World<*, *>, args: LuaThread.ArgStack): CompletableFuture<TileDamageResult> {
	val positions = args.readTableValues { getVector2i(it) ?: throw IllegalArgumentException("Positions table contains invalid positions") }
	val isBackground = args.nextBoolean()

	val sourcePosition = args.nextVector2d()
	val damageType = TileDamageType.entries.valueOf(args.nextString())
	val damage = args.nextDouble()
	val harvestLevel = args.nextOptionalLong()?.toInt() ?: 999
	val sourceEntity = self.entities[args.nextOptionalLong()?.toInt() ?: 0]

	return self.damageTiles(positions, isBackground, sourcePosition, TileDamage(damageType, damage, harvestLevel), sourceEntity)
}

private fun damageTileAreaImpl(self: World<*, *>, args: LuaThread.ArgStack): CompletableFuture<TileDamageResult> {
	val center = args.nextVector2i()
	val radius = args.nextDouble()
	val isBackground = args.nextBoolean()

	val sourcePosition = args.nextVector2d()
	val damageType = TileDamageType.entries.valueOf(args.nextString())
	val damage = args.nextDouble()
	val harvestLevel = args.nextOptionalLong()?.toInt() ?: 999
	val sourceEntity = self.entities[args.nextOptionalLong()?.toInt() ?: 0]

	return self.damageTiles(tileAreaBrush(center, radius), isBackground, sourcePosition, TileDamage(damageType, damage, harvestLevel), sourceEntity)
}

private fun placeMaterialImpl(self: World<*, *>, args: LuaThread.ArgStack): CompletableFuture<List<Vector2i>> {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val material = Registries.tiles.getOrThrow(args.nextString())
	val hueShift: Float? = args.nextOptionalDouble()?.let { it.toFloat() / 255f * 360f }
	val allowOverlap = args.nextBoolean()

	val action = TileModification.PlaceMaterial(isBackground, material.ref, hueShift)

	// original engine return here doesn't make any sense whatsoever
	return self.applyTileModifications(listOf(pos to action), allowOverlap, false).thenApply { it.map { it.first } }
}

private fun placeModImpl(self: World<*, *>, args: LuaThread.ArgStack): CompletableFuture<List<Vector2i>> {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val material = Registries.tileModifiers.getOrThrow(args.nextString())
	val hueShift: Float? = args.nextOptionalDouble()?.let { it.toFloat() / 255f * 360f }
	val allowOverlap = args.nextBoolean()

	val action = TileModification.PlaceModifier(isBackground, material.ref, hueShift)

	// original engine return here doesn't make any sense whatsoever
	return self.applyTileModifications(listOf(pos to action), allowOverlap, false).thenApply { it.map { it.first } }
}

private fun breathable(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.chunkMap.isBreathable(args.nextVector2d()))
	return 1
}

private fun underground(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.undergroundLevel >= args.nextVector2d().y)
	return 1
}

private fun material(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()

	val tile = self.getCell(pos).tile(isBackground)

	if (tile.material.isNullTile) {
		return 0
	} else if (tile.material.isEmptyTile) {
		args.lua.push(false)
		return 1
	} else {
		args.lua.push(tile.material.key)
		return 1
	}
}

private fun mod(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val tile = self.getCell(pos).tile(isBackground)

	if (tile.modifier.isNotEmptyModifier) {
		args.lua.push(tile.modifier.key)
		return 1
	}

	return 0
}

private fun materialHueShift(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val tile = self.getCell(pos).tile(isBackground)

	args.lua.push(tile.hueShift)
	return 1
}

private fun modHueShift(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val tile = self.getCell(pos).tile(isBackground)

	args.lua.push(tile.modifierHueShift)
	return 1
}

private fun materialColor(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val tile = self.getCell(pos).tile(isBackground)

	args.lua.push(tile.color.ordinal.toLong())
	return 1
}

private fun materialColorName(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val tile = self.getCell(pos).tile(isBackground)

	args.lua.push(tile.color.jsonName)
	return 1
}

private fun setMaterialColorNumber(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val getColor = args.nextInt()
	val color = TileColor.entries.getOrNull(getColor) ?: throw IllegalArgumentException("invalid color: $getColor")

	val cell = self.getCell(pos).mutable()
	cell.tile(isBackground).color = color
	args.lua.push(self.setCell(pos, cell))
	return 1
}

private fun setMaterialColorName(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val pos = args.nextVector2i()
	val isBackground = args.nextBoolean()
	val getColor = args.nextString()
	val color = TileColor.entries.valueOf(getColor)

	val cell = self.getCell(pos).mutable()
	cell.tile(isBackground).color = color
	args.lua.push(self.setCell(pos, cell))
	return 1
}

private fun oceanLevel(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(self.template.cellInfo(args.nextVector2d()).oceanLiquidLevel.toLong())
	return 1
}

private fun environmentStatusEffects(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val effects = self.environmentStatusEffects(args.nextVector2d())

	args.lua.pushTable(effects.size)

	for ((i, effect) in effects.withIndex()) {
		args.lua.push(i + 1L)
		args.lua.push(Starbound.gson.toJsonTree(effect))
		args.lua.setTableValue()
	}

	return 1
}

private fun weatherStatusEffects(self: World<*, *>, args: LuaThread.ArgStack): Int {
	val effects = self.weatherStatusEffects(args.nextVector2d())

	args.lua.pushTable(effects.size)

	for ((i, effect) in effects.withIndex()) {
		args.lua.push(i + 1L)
		args.lua.push(Starbound.gson.toJsonTree(effect))
		args.lua.setTableValue()
	}

	return 1
}

private fun damageTiles(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(damageTilesImpl(self, args).getNow(TileDamageResult.NONE) != TileDamageResult.NONE)
	return 1
}

private fun damageTilesPromise(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = damageTilesImpl(self, args).thenApply { it.jsonName },
			isLocal = false,
			handler = {
				push(it)
				1
			}
		)
	)

	return 1
}

private fun damageTileArea(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(damageTileAreaImpl(self, args).getNow(TileDamageResult.NONE) != TileDamageResult.NONE)
	return 1
}

private fun damageTileAreaPromise(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = damageTileAreaImpl(self, args).thenApply { it.jsonName },
			isLocal = false,
			handler = {
				push(it)
				1
			}
		)
	)

	return 1
}

private fun placeMaterial(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(placeMaterialImpl(self, args).getNow(listOf()).isEmpty())
	return 1
}

private fun placeMaterialPromise(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = placeMaterialImpl(self, args),
			isLocal = false,
			handler = {
				pushTable(it.size)

				for ((i, pos) in it.withIndex()) {
					push(i + 1L)
					push(pos)
					setTableValue()
				}

				1
			}
		)
	)

	return 1
}

private fun placeMod(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(placeModImpl(self, args).getNow(listOf()).isEmpty())
	return 1
}

private fun placeModPromise(self: World<*, *>, args: LuaThread.ArgStack): Int {
	args.lua.push(
		LuaFuture(
			future = placeModImpl(self, args),
			isLocal = false,
			handler = {
				pushTable(it.size)

				for ((i, pos) in it.withIndex()) {
					push(i + 1L)
					push(pos)
					setTableValue()
				}

				1
			}
		)
	)
	return 1
}

fun provideWorldEnvironmentalBindings(self: World<*, *>, lua: LuaThread) {
	lua.setTableValueToStub("lightLevel")
	lua.setTableValueToStub("windLevel")

	lua.pushBinding(self, "breathable", ::breathable)
	lua.pushBinding(self, "underground", ::underground)

	lua.pushBinding(self, "material", ::material)
	lua.pushBinding(self, "mod", ::mod)
	lua.pushBinding(self, "materialHueShift", ::materialHueShift)
	lua.pushBinding(self, "modHueShift", ::modHueShift)
	lua.pushBinding(self, "materialColor", ::materialColor)
	lua.pushBinding(self, "materialColorName", ::materialColorName)
	lua.pushBinding(self, "setMaterialColorNumber", ::setMaterialColorNumber)
	lua.pushBinding(self, "setMaterialColorName", ::setMaterialColorName)

	lua.pushBinding(self, "oceanLevel", ::oceanLevel)
	lua.pushBinding(self, "environmentStatusEffects", ::environmentStatusEffects)
	lua.pushBinding(self, "weatherStatusEffects", ::weatherStatusEffects)

	lua.pushBinding(self, "damageTiles", ::damageTiles)
	lua.pushBinding(self, "damageTilesPromise", ::damageTilesPromise)

	lua.pushBinding(self, "damageTileArea", ::damageTileArea)
	lua.pushBinding(self, "damageTileAreaPromise", ::damageTileAreaPromise)

	lua.pushBinding(self, "placeMaterial", ::placeMaterial)
	lua.pushBinding(self, "placeMaterialPromise", ::placeMaterialPromise)

	lua.pushBinding(self, "placeMod", ::placeMod)
	lua.pushBinding(self, "placeModPromise", ::placeModPromise)
}
