package ru.dbotthepony.kstarbound.util

import it.unimi.dsi.fastutil.objects.ObjectArrayList
import kotlinx.coroutines.CancellationException
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.SupervisorJob
import kotlinx.coroutines.asCoroutineDispatcher
import kotlinx.coroutines.cancel
import org.apache.logging.log4j.LogManager
import java.util.PriorityQueue
import java.util.concurrent.Callable
import java.util.concurrent.CompletableFuture
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.Delayed
import java.util.concurrent.Future
import java.util.concurrent.FutureTask
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.RejectedExecutionException
import java.util.concurrent.ScheduledExecutorService
import java.util.concurrent.ScheduledFuture
import java.util.concurrent.TimeUnit
import java.util.concurrent.locks.LockSupport
import java.util.function.Supplier
import kotlin.math.absoluteValue

// I tried to make use of Netty's event loops, but they seem to be a bit overcomplicated
// if you try to use them by yourself :(
// so I made my own
open class BlockableEventLoop(name: String) : Thread(name), ScheduledExecutorService {
	private class ScheduledTask<T>(callable: Callable<T>, var executeAt: Long, val repeat: Boolean, val timeDelay: Long, val isFixedDelay: Boolean, generation: Long) : FutureTask<T>(callable), ScheduledFuture<T> {
		var generation = generation
			private set

		override fun compareTo(other: Delayed): Int {
			if (other is ScheduledTask<*>)
				return executeAt.compareTo(other.executeAt)

			return getDelay(TimeUnit.NANOSECONDS).compareTo(other.getDelay(TimeUnit.NANOSECONDS))
		}

		override fun getDelay(unit: TimeUnit): Long {
			return executeAt - System.nanoTime()
		}

		fun shouldExecuteNow(): Boolean {
			return executeAt <= System.nanoTime()
		}

		fun shouldEnqueue(isShutdown: Boolean, generation: Long): Boolean {
			if (isShutdown || executeAt <= System.nanoTime())
				return perform(isShutdown, generation)

			return true
		}

		fun perform(isShutdown: Boolean, generation: Long): Boolean {
			if (repeat) {
				if (isFixedDelay) {
					// fixed delay

					val deadlineMargin = executeAt - System.nanoTime()

					if (deadlineMargin <= -5_000_000_000L) {
						if (!IS_IN_IDE) // since this will get spammed if debugging using breakpoints
							LOGGER.warn("Event loop missed scheduled deadline by ${-deadlineMargin / 1_000_000L} milliseconds")
					}

					val result = runAndReset()
					executeAt = System.nanoTime() + timeDelay
					this.generation = generation
					return result && !isShutdown
				} else {
					// fixed rate
					val timeBefore = System.nanoTime()
					var deadlineMargin = executeAt - System.nanoTime()

					if (deadlineMargin <= -5_000_000_000L) {
						if (!IS_IN_IDE) // since this will get spammed if debugging using breakpoints
							LOGGER.warn("Event loop missed scheduled deadline by ${-deadlineMargin / 1_000_000L} milliseconds, clamping to 5 seconds")

						deadlineMargin = -5_000_000_000L
					}

					val result = runAndReset()
					val now = System.nanoTime()
					executeAt = now + timeDelay + deadlineMargin - (now - timeBefore)
					this.generation = generation
					return result && !isShutdown
				}
			} else {
				run()
				return false
			}
		}
	}

	private class TaskPair<T>(val future: CompletableFuture<T>, var supplier: Callable<T>?)

	private var generation = 0L
	private var extraBudget = 0L
	private val eventQueue = ConcurrentLinkedQueue<TaskPair<*>>()
	private val scheduledQueue = PriorityQueue<ScheduledTask<*>>()
	val coroutines = asCoroutineDispatcher()
	val scope = CoroutineScope(coroutines + SupervisorJob())

	init {
		priority = MAX_PRIORITY
	}

	private fun nextDeadline(): Long {
		if (isShutdown || eventQueue.isNotEmpty())
			return 0L

		val poll = scheduledQueue.peek()

		if (poll == null) {
			return Long.MAX_VALUE
		} else {
			return poll.executeAt - System.nanoTime()
		}
	}

	@Volatile
	private var isShutdown = false
	private var isRunning = true

	private fun eventLoopIteration(): Boolean {
		var executedAnything = false
		// poll queued event first, so we guarantee at least one is executed even if main event loop is very busy
		var next = eventQueue.poll()

		while (next != null) {
			executedAnything = true
			val time = System.nanoTime()

			try {
				val callable = next.supplier

				if (callable != null) {
					(next.future as CompletableFuture<Any?>).complete(callable.call())
				}
			} catch (err: Throwable) {
				LOGGER.error("Exception executing scheduled task", err)

				try {
					next.future.completeExceptionally(err)
				} catch (err: Throwable) {
					LOGGER.error("Caught an exception while propagating CompletableFuture to completeExceptionally stage", err)
				}
			}

			if (extraBudget > 0L) {
				// event loop is overloaded, keep executing queued tasks until we empty execution time budget
				extraBudget -= System.nanoTime() - time
				next = eventQueue.poll()
			} else if (scheduledQueue.isEmpty() || !scheduledQueue.peek()!!.shouldExecuteNow())
				// keep executing queued tasks until we hit scheduled task deadline
				next = eventQueue.poll()
			else
				next = null
		}

		if (scheduledQueue.isNotEmpty()) {
			val time = System.nanoTime()
			val executed = ObjectArrayList<ScheduledTask<*>>(4)
			// first, poll all due events into local queue, and prioritize events by FIFO strategy based on how long ago they were scheduled
			val queue = PriorityQueue<ScheduledTask<*>>(Comparator { o1, o2 -> o1.generation.compareTo(o2.generation) })

			while (scheduledQueue.isNotEmpty() && (isShutdown || scheduledQueue.peek()!!.shouldExecuteNow())) {
				queue.add(scheduledQueue.poll()!!)
			}

			executedAnything = executedAnything || queue.isNotEmpty()

			// then, execute collected events in proper order
			while (queue.isNotEmpty()) {
				val poll = queue.poll()!!

				if (poll.perform(isShutdown, generation++)) {
					executed.add(poll)
				}
			}

			scheduledQueue.addAll(executed)

			if (scheduledQueue.isNotEmpty() && !isShutdown && scheduledQueue.peek()!!.shouldExecuteNow()) {
				// event loop is overloaded, interleave queued events with scheduled ones
				extraBudget = System.nanoTime() - time
			} else {
				// no overload, normal queued events execution
				extraBudget = 0L
			}
		} else {
			extraBudget = 0L
		}

		return executedAnything
	}

	final override fun run() {
		while (isRunning) {
			LockSupport.parkNanos(nextDeadline())
			eventLoopIteration()

			if (isShutdown && isRunning) {
				while (eventLoopIteration()) {}

				try {
					scope.cancel(CancellationException("EventLoop shut down"))
					performShutdown()
				} catch (err: Throwable) {
					LOGGER.fatal("Exception shutting down $name")
					return
				} finally {
					isRunning = false
				}
			}
		}

		LOGGER.info("Thread $name stopped gracefully")
	}

	final override fun execute(command: Runnable) {
		if (currentThread() === this) {
			command.run()
		} else {
			if (isShutdown)
				throw RejectedExecutionException("EventLoop is shutting down")

			val future = CompletableFuture<Unit>()
			val pair = TaskPair(future) { command.run() }

			future.exceptionally {
				pair.supplier = null
			}

			eventQueue.add(pair)
			LockSupport.unpark(this)
		}
	}

	final override fun <T> submit(task: Callable<T>): CompletableFuture<T> {
		if (currentThread() === this) {
			try {
				return CompletableFuture.completedFuture(task.call())
			} catch (err: Throwable) {
				return CompletableFuture.failedFuture(err)
			}
		} else {
			if (isShutdown)
				throw RejectedExecutionException("EventLoop is shutting down")

			val future = CompletableFuture<T>()
			val pair = TaskPair(future, task)

			future.exceptionally {
				pair.supplier = null
				null
			}

			eventQueue.add(pair)
			LockSupport.unpark(this)
			return future
		}
	}

	fun <T> supplyAsync(task: Supplier<T>): CompletableFuture<T> {
		return submit(task::get) as CompletableFuture<T>
	}

	fun <T> supplyAsync(task: () -> T): CompletableFuture<T> {
		return submit(task::invoke) as CompletableFuture<T>
	}

	fun ensureSameThread() {
		check(this === currentThread()) { "Performing non-threadsafe operation outside of event loop thread $this" }
	}

	fun isSameThread() = this === currentThread()

	final override fun submit(task: Runnable): CompletableFuture<*> {
		if (currentThread() === this) {
			try {
				return CompletableFuture.completedFuture(task.run())
			} catch (err: Throwable) {
				return CompletableFuture.failedFuture<Unit>(err)
			}
		} else {
			if (isShutdown)
				throw RejectedExecutionException("EventLoop is shutting down")

			val future = CompletableFuture<Unit>()
			val pair = TaskPair(future) { task.run() }

			future.exceptionally {
				pair.supplier = null
			}

			eventQueue.add(pair)
			LockSupport.unpark(this)
			return future
		}
	}

	final override fun <T> submit(task: Runnable, result: T): CompletableFuture<T> {
		if (currentThread() === this) {
			try {
				task.run()
				return CompletableFuture.completedFuture(result)
			} catch (err: Throwable) {
				return CompletableFuture.failedFuture(err)
			}
		} else {
			if (isShutdown)
				throw RejectedExecutionException("EventLoop is shutting down")

			val future = CompletableFuture<T>()
			val pair = TaskPair(future) { task.run(); result }

			future.exceptionally {
				pair.supplier = null
				null
			}

			eventQueue.add(pair)
			LockSupport.unpark(this)
			return future
		}
	}

	override fun close() {
		// do nothing
	}

	final override fun <T> invokeAll(tasks: Collection<Callable<T>>): List<Future<T>> {
		return tasks.map { submit(it) }
	}

	final override fun <T> invokeAll(tasks: Collection<Callable<T>>, timeout: Long, unit: TimeUnit): List<Future<T>> {
		val futures = tasks.map { submit(it) }
		CompletableFuture.allOf(*futures.toTypedArray()).get(timeout, unit)
		return futures
	}

	final override fun <T> invokeAny(tasks: Collection<Callable<T>>): T {
		return submit(tasks.first()).get()
	}

	final override fun <T> invokeAny(tasks: Collection<Callable<T>>, timeout: Long, unit: TimeUnit): T {
		return submit(tasks.first()).get(timeout, unit)
	}

	final override fun shutdown() {
		if (!isShutdown) {
			isShutdown = true

			if (currentThread() === this || state == State.NEW) {
				LOGGER.info("Shutdown initiated")
				while (eventLoopIteration()) {}

				while (scheduledQueue.isNotEmpty()) {
					val remove = scheduledQueue.remove()

					try {
						remove.cancel(false)
					} catch (err: Throwable) {
						LOGGER.warn("Caught exception while cancelling future during event loop shutdown", err)
					}
				}

				try {
					scope.cancel(CancellationException("EventLoop shut down"))
					performShutdown()
				} catch (err: Throwable) {
					LOGGER.fatal("Exception shutting down $name", err)
				} finally {
					isRunning = false
				}
			} else {
				LOGGER.info("$name shutdown initiated")
				// wake up thread
				LockSupport.unpark(this)
			}
		}
	}

	protected open fun performShutdown() {

	}

	final override fun shutdownNow(): List<Runnable> {
		if (!isShutdown) {
			isShutdown = true

			if (currentThread() === this) {
				while (eventQueue.isNotEmpty()) {
					val remove = eventQueue.remove()

					try {
						remove.future.cancel(false)
					} catch (err: Throwable) {
						LOGGER.warn("Caught exception while cancelling future during event loop shutdown", err)
					}
				}

				while (scheduledQueue.isNotEmpty()) {
					val remove = scheduledQueue.remove()

					try {
						remove.cancel(false)
					} catch (err: Throwable) {
						LOGGER.warn("Caught exception while cancelling future during event loop shutdown", err)
					}
				}

				try {
					scope.cancel(CancellationException("EventLoop shut down"))
					performShutdown()
				} catch (err: Throwable) {
					LOGGER.fatal("Exception shutting down $name")
				} finally {
					isRunning = false
				}
			} else {
				// wake up thread
				LockSupport.unpark(this)
			}
		}

		return emptyList()
	}

	final override fun awaitTermination(timeout: Long, unit: TimeUnit): Boolean {
		// lazy wait loop
		var budget = TimeUnit.NANOSECONDS.convert(timeout, unit)
		var origin = System.nanoTime()

		while (budget > 0L && isRunning) {
			val new = System.nanoTime()
			budget -= new - origin
			origin = new
			LockSupport.parkNanos(budget.coerceAtMost(500_000L))

			if (interrupted()) {
				return !isRunning
			}
		}

		return !isRunning
	}

	final override fun isShutdown(): Boolean {
		return isShutdown
	}

	final override fun isTerminated(): Boolean {
		return !isRunning
	}

	final override fun schedule(command: Runnable, delay: Long, unit: TimeUnit): ScheduledFuture<*> {
		val task = ScheduledTask({ command.run() }, System.nanoTime() + TimeUnit.NANOSECONDS.convert(delay, unit), false, 0L, false, generation++)

		execute {
			if (task.shouldEnqueue(isShutdown, generation++))
				scheduledQueue.add(task)
		}

		return task
	}

	final override fun <V> schedule(callable: Callable<V>, delay: Long, unit: TimeUnit): ScheduledFuture<V> {
		val task = ScheduledTask(callable, System.nanoTime() + TimeUnit.NANOSECONDS.convert(delay, unit), false, 0L, false, generation++)

		execute {
			if (task.shouldEnqueue(isShutdown, generation++))
				scheduledQueue.add(task)
		}

		return task
	}

	final override fun scheduleAtFixedRate(
		command: Runnable,
		initialDelay: Long,
		period: Long,
		unit: TimeUnit
	): ScheduledFuture<*> {
		val task = ScheduledTask({ command.run() }, System.nanoTime() + TimeUnit.NANOSECONDS.convert(initialDelay, unit), true, TimeUnit.NANOSECONDS.convert(period, unit), false, generation++)

		execute {
			if (task.shouldEnqueue(isShutdown, generation++))
				scheduledQueue.add(task)
		}

		return task
	}

	final override fun scheduleWithFixedDelay(
		command: Runnable,
		initialDelay: Long,
		delay: Long,
		unit: TimeUnit
	): ScheduledFuture<*> {
		val task = ScheduledTask({ command.run() }, System.nanoTime() + TimeUnit.NANOSECONDS.convert(initialDelay, unit), true, TimeUnit.NANOSECONDS.convert(delay, unit), true, generation++)

		execute {
			if (task.shouldEnqueue(isShutdown, generation++))
				scheduledQueue.add(task)
		}

		return task
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
		private val IS_IN_IDE = java.lang.management.ManagementFactory.getRuntimeMXBean().inputArguments.toString().contains("-agentlib:jdwp")
	}
}
