package ru.dbotthepony.kstarbound.world.entities

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import org.classdump.luna.ByteString
import org.classdump.luna.Table
import ru.dbotthepony.kommons.util.IStruct2d
import ru.dbotthepony.kstarbound.defs.DamageNotification
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.defs.Drawable
import ru.dbotthepony.kstarbound.defs.InteractAction
import ru.dbotthepony.kstarbound.defs.InteractRequest
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.lua.LuaEnvironment
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.get
import ru.dbotthepony.kstarbound.lua.tableMapOf
import ru.dbotthepony.kstarbound.lua.toJsonFromLua
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.network.packets.DamageNotificationPacket
import ru.dbotthepony.kstarbound.network.packets.DamageRequestPacket
import ru.dbotthepony.kstarbound.network.packets.HitRequestPacket
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.api.InteractiveEntity
import ru.dbotthepony.kstarbound.world.entities.api.ScriptedEntity

/**
 * Monsters, NPCs, Players
 */
abstract class ActorEntity : DynamicEntity(), InteractiveEntity, ScriptedEntity {
	final override val movement: ActorMovementController = ActorMovementController()
	abstract val statusController: StatusController

	// ticked manually in final classes, to ensure proper order of operations
	val effects = EffectEmitter(this)

	abstract val lua: LuaEnvironment

	override fun move(delta: Double) {
		statusController.applyMovementControls()
		super.move(delta)
	}

	// uint8_t
	enum class DamageBarType(override val jsonName: String) : IStringSerializable {
		DEFAULT("Default"), NONE("None"), SPECIAL("Special")
	}

	var health: Double
		get() = statusController.resources["health"]!!.value
		set(value) { statusController.resources["health"]!!.value = value }
	val maxHealth: Double
		get() = statusController.resources["health"]!!.maxValue!!

	val isDead: Boolean
		get() = health <= 0.0

	val isInvulnerable: Boolean
		get() = statusController.statPositive("invulnerable")

	abstract val damageBarType: DamageBarType

	override fun onJoinWorld(world: World<*, *>) {
		super.onJoinWorld(world)
		statusController.init()
	}

	override fun tick(delta: Double) {
		super.tick(delta)

		if (isLocal)
			statusController.tick(delta)
		else
			statusController.tickRemote(delta)
	}

	override fun hitOther(damage: HitRequestPacket) {
		super.hitOther(damage)
		statusController.notifyHitDealt(damage.target, damage.request)
	}

	override fun damagedOther(notification: DamageNotificationPacket) {
		super.damagedOther(notification)
		statusController.notifyDamageDealt(notification.notification)
	}

	enum class PortraitMode(override val jsonName: String) : IStringSerializable {
		HEAD("head"),
		BUST("bust"),
		FULL("full"),
		FULL_NEUTRAL("fullneutral"),
		FULL_NUDE("fullnude"),
		FULL_NEUTRAL_NUDE("fullneutralnude");
	}

	@Suppress("NAME_SHADOWING")
	fun toAbsolutePosition(pos: Vector2d): Vector2d {
		var pos = pos

		if (movement.facingDirection == Direction.LEFT) {
			pos = pos.copy(x = -pos.x)
		}

		if (movement.rotation != 0.0) {
			pos = pos.rotate(movement.rotation)
		}

		return movement.position + pos
	}

	open fun portrait(mode: PortraitMode): List<Drawable> {
		return emptyList()
	}

	override fun callScript(fnName: String, arguments: JsonArray): JsonElement {
		//require(isLocal) { "Calling script on remote entity" }
		//return lua.invokeGlobal(fnName, *arguments)
		TODO()
	}

	override fun evalScript(code: String): Array<Any?> {
		require(isLocal) { "Calling script on remote entity" }
		return lua.eval(code)
	}

	override fun interact(request: InteractRequest): InteractAction {
		val result = lua.invokeGlobal("interact", lua.tableMapOf(
			"sourceId" to request.source,
			"sourcePosition" to lua.from(request.sourcePos)
		))

		if (result.isEmpty() || result[0] == null)
			return InteractAction.NONE

		val value = result[0]

		if (value is ByteString)
			return InteractAction(value.decode(), entityID)

		value as Table
		return InteractAction((value[1L] as ByteString).decode(), entityID, toJsonFromLua(value[2L]))
	}
}
