package ru.dbotthepony.kstarbound.world.entities

import it.unimi.dsi.fastutil.objects.ObjectArrayList
import ru.dbotthepony.kommons.util.getValue
import ru.dbotthepony.kommons.util.setValue
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.DamageSource
import ru.dbotthepony.kstarbound.defs.actor.HumanoidConfig
import ru.dbotthepony.kstarbound.defs.actor.HumanoidEmote
import ru.dbotthepony.kstarbound.defs.actor.HumanoidIdentity
import ru.dbotthepony.kstarbound.defs.actor.MoveControlType
import ru.dbotthepony.kstarbound.defs.actor.PersistentStatusEffect
import ru.dbotthepony.kstarbound.defs.item.ItemDescriptor
import ru.dbotthepony.kstarbound.io.NullableBinaryStringCodec
import ru.dbotthepony.kstarbound.item.ItemStack
import ru.dbotthepony.kstarbound.item.ToolItem
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.math.Interpolator
import ru.dbotthepony.kstarbound.network.syncher.NetworkedElement
import ru.dbotthepony.kstarbound.network.syncher.NetworkedGroup
import ru.dbotthepony.kstarbound.network.syncher.networkedBoolean
import ru.dbotthepony.kstarbound.network.syncher.networkedData
import ru.dbotthepony.kstarbound.network.syncher.networkedEnum
import ru.dbotthepony.kstarbound.network.syncher.networkedFixedPoint
import ru.dbotthepony.kstarbound.network.syncher.networkedItem
import ru.dbotthepony.kstarbound.network.syncher.networkedStatefulItem
import ru.dbotthepony.kstarbound.util.GameTimer
import ru.dbotthepony.kstarbound.util.random.nextRange
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.entities.api.LoungeableEntity
import kotlin.reflect.KMutableProperty1

/**
 * Players and NPCs
 */
abstract class HumanoidActorEntity : ActorEntity() {
	abstract var xAimPosition: Double
	abstract var yAimPosition: Double

	var energy: Double
		get() = statusController.resources["energy"]!!.value
		set(value) { statusController.resources["energy"]!!.value = value }
	val maxEnergy: Double
		get() = statusController.resources["energy"]!!.maxValue!!

	var aimPosition: Vector2d
		get() = Vector2d(xAimPosition, yAimPosition)
		set(value) {
			xAimPosition = value.x
			yAimPosition = value.y
		}

	enum class ItemSlot(override val jsonName: String, val property: KMutableProperty1<HumanoidActorEntity, ItemStack>) : IStringSerializable {
		HEAD("head", HumanoidActorEntity::headItem),
		HEAD_COSMETIC("headcosmetic", HumanoidActorEntity::headCosmeticItem),
		CHEST("chest", HumanoidActorEntity::chestItem),
		CHEST_COSMETIC("chestcosmetic", HumanoidActorEntity::chestCosmeticItem),
		LEGS("legs", HumanoidActorEntity::legsItem),
		LEGS_COSMETIC("legscosmetic", HumanoidActorEntity::legsCosmeticItem),
		BACK("back", HumanoidActorEntity::backItem),
		BACK_COSMETIC("backcosmetic", HumanoidActorEntity::backCosmeticItem),
		PRIMARY("primary", HumanoidActorEntity::primaryHandItem),
		ALT("alt", HumanoidActorEntity::secondaryHandItem) {
			override fun match(name: String): Boolean {
				return super.match(name) || name == "secondary"
			}
		};
	}

	fun setItem(slot: ItemSlot, item: ItemStack) {
		slot.property.set(this, item)
	}

	fun setItem(slot: String, item: ItemStack): Boolean {
		val lower = slot.lowercase()
		setItem(ItemSlot.entries.firstOrNull { it.match(lower) } ?: return false, item.copy())
		return true
	}

	fun setItem(slot: String, item: ItemDescriptor): Boolean {
		val lower = slot.lowercase()
		setItem(ItemSlot.entries.firstOrNull { it.match(lower) } ?: return false, item.build())
		return true
	}

	fun getItem(slot: ItemSlot): ItemStack {
		return slot.property.get(this)
	}

	// it makes no sense to split ToolUser' logic into separate class
	// StarToolUser
	protected val toolsNetworkGroup = NetworkedGroup()

	var primaryHandItem by toolsNetworkGroup.add(networkedStatefulItem())
	var secondaryHandItem by toolsNetworkGroup.add(networkedStatefulItem())
	var primaryFireTimerNetState by toolsNetworkGroup.add(networkedFixedPoint(Starbound.TIMESTEP).also { it.interpolator = ToolFiringInterpolator })
	var secondaryFireTimerNetState by toolsNetworkGroup.add(networkedFixedPoint(Starbound.TIMESTEP).also { it.interpolator = ToolFiringInterpolator })
	var primaryTimeFiringNetState by toolsNetworkGroup.add(networkedFixedPoint(Starbound.TIMESTEP).also { it.interpolator = ToolFiringInterpolator })
	var secondaryTimeFiringNetState by toolsNetworkGroup.add(networkedFixedPoint(Starbound.TIMESTEP).also { it.interpolator = ToolFiringInterpolator })
	var primaryItemActive by toolsNetworkGroup.add(networkedBoolean())
	var secondaryItemActive by toolsNetworkGroup.add(networkedBoolean())

	fun beginPrimaryFire() {
		TODO()
	}

	fun beginSecondaryFire() {
		TODO()
	}

	fun endPrimaryFire() {
		TODO()
	}

	fun endSecondaryFire() {
		TODO()
	}

	private object ToolFiringInterpolator : Interpolator {
		override fun interpolate(t: Double, a: Double, b: Double): Double {
			return if (a >= b) b else Interpolator.Linear.interpolate(t, a, b)
		}
	}

	@JsonFactory
	data class ToolSerializedData(
		val primaryHandItem: ItemDescriptor = ItemDescriptor.EMPTY,
		// backwards compatible name
		val altHandItem: ItemDescriptor = ItemDescriptor.EMPTY,

		// original engine will ignore these
		val primaryFireTimerNetState: Double = 0.0,
		val secondaryFireTimerNetState: Double = 0.0,
		val primaryTimeFiringNetState: Double = 0.0,
		val secondaryTimeFiringNetState: Double = 0.0,
		val primaryItemActive: Boolean = false,
		val secondaryItemActive: Boolean = false,
	)

	// to match semantics of original engine, and also because these are called *only* by NPCs
	// :upside_down:
	protected fun serializeToolUser(): ToolSerializedData {
		return ToolSerializedData(
			primaryHandItem = primaryHandItem.createDescriptor(),
			altHandItem = secondaryHandItem.createDescriptor(),
			primaryFireTimerNetState = primaryFireTimerNetState,
			secondaryFireTimerNetState = secondaryFireTimerNetState,
			primaryTimeFiringNetState = primaryTimeFiringNetState,
			secondaryTimeFiringNetState = secondaryTimeFiringNetState,
			primaryItemActive = primaryItemActive,
			secondaryItemActive = secondaryItemActive,
		)
	}

	protected fun deserializeToolUser(data: ToolSerializedData) {
		primaryHandItem = data.primaryHandItem.build()
		secondaryHandItem = data.altHandItem.build()
		primaryFireTimerNetState = data.primaryFireTimerNetState
		secondaryFireTimerNetState = data.secondaryFireTimerNetState
		primaryTimeFiringNetState = data.primaryTimeFiringNetState
		secondaryTimeFiringNetState = data.secondaryTimeFiringNetState
		primaryItemActive = data.primaryItemActive
		secondaryItemActive = data.secondaryItemActive
	}

	protected fun queryShieldHit(damage: DamageSource): Boolean {
		TODO()
	}

	// StarArmorUser
	protected val armorNetworkGroup = NetworkedGroup()

	var headItem by armorNetworkGroup.add(networkedItem())
	var chestItem by armorNetworkGroup.add(networkedItem())
	var legsItem by armorNetworkGroup.add(networkedItem())
	var backItem by armorNetworkGroup.add(networkedItem())
	var headCosmeticItem by armorNetworkGroup.add(networkedItem())
	var chestCosmeticItem by armorNetworkGroup.add(networkedItem())
	var legsCosmeticItem by armorNetworkGroup.add(networkedItem())
	var backCosmeticItem by armorNetworkGroup.add(networkedItem())

	@JsonFactory
	data class ArmorSerializedData(
		val headItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val chestItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val legsItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val backItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val headCosmeticItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val chestCosmeticItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val legsCosmeticItem: ItemDescriptor = ItemDescriptor.EMPTY,
		val backCosmeticItem: ItemDescriptor = ItemDescriptor.EMPTY,
	)

	// to match semantics of original engine, and also because these are called *only* by NPCs
	// :upside_down:
	protected fun serializeArmorUser(): ArmorSerializedData {
		return ArmorSerializedData(
			headItem = headItem.createDescriptor(),
			chestItem = chestItem.createDescriptor(),
			legsItem = legsItem.createDescriptor(),
			backItem = backItem.createDescriptor(),
			headCosmeticItem = headCosmeticItem.createDescriptor(),
			chestCosmeticItem = chestCosmeticItem.createDescriptor(),
			legsCosmeticItem = legsCosmeticItem.createDescriptor(),
			backCosmeticItem = backCosmeticItem.createDescriptor(),
		)
	}

	protected fun deserializeArmorUser(data: ArmorSerializedData) {
		headItem = data.headItem.build()
		chestItem = data.chestItem.build()
		legsItem = data.legsItem.build()
		backItem = data.backItem.build()
		headCosmeticItem = data.headCosmeticItem.build()
		chestCosmeticItem = data.chestCosmeticItem.build()
		legsCosmeticItem = data.legsCosmeticItem.build()
		backCosmeticItem = data.backCosmeticItem.build()
	}

	// ---

	override val damageSources: List<DamageSource> get() {
		val damageSources = ObjectArrayList<DamageSource>()
		val primaryHandItem = primaryHandItem
		val secondaryHandItem = secondaryHandItem

		if (primaryHandItem is ToolItem) damageSources.addAll(primaryHandItem.damageSources)
		if (secondaryHandItem is ToolItem) damageSources.addAll(secondaryHandItem.damageSources)

		return damageSources
	}

	// StarHumanoid
	abstract val humanoidIdentity: HumanoidIdentity
	abstract val humanoidConfig: HumanoidConfig

	final override val mouthPosition: Vector2d
		get() = movement.getAbsolutePosition(humanoidConfig.mouthOffset)

	final override val feetPosition: Vector2d
		get() = movement.getAbsolutePosition(humanoidConfig.feetOffset)

	// TODO: Original engine does not account for rotation here, was this intended, or did we fix a bug?
	val headArmorPosition: Vector2d
		get() = movement.getAbsolutePosition(humanoidConfig.headArmorOffset)

	// TODO: Original engine does not account for rotation here, was this intended, or did we fix a bug?
	val chestArmorPosition: Vector2d
		get() = movement.getAbsolutePosition(humanoidConfig.chestArmorOffset)

	// TODO: Original engine does not account for rotation here, was this intended, or did we fix a bug?
	val legsArmorPosition: Vector2d
		get() = movement.getAbsolutePosition(humanoidConfig.legsArmorOffset)

	// TODO: Original engine does not account for rotation here, was this intended, or did we fix a bug?
	val backArmorPosition: Vector2d
		get() = movement.getAbsolutePosition(humanoidConfig.backArmorOffset)

	inner class Hand {
		var angle = 0.0
	}

	var isTwoHanded = false
	val primaryHand = Hand()
	val secondaryHand = Hand()

	var isMovingBackwards = false

	enum class HumanoidState(override val jsonName: String) : IStringSerializable {
		IDLE("idle"), // 1 idle frame
		WALK("walk"), // 8 walking frames
		RUN("run"), // 8 run frames
		JUMP("jump"), // 4 jump frames
		FALL("fall"), // 4 fall frames
		SWIM("swim"), // 7 swim frames
		SWIM_IDLE("swimIdle"), // 2 swim idle frame
		DUCK("duck"), // 1 ducking frame
		SIT("sit"), // 1 sitting frame
		LAY("lay"); // 1 laying frame
	}

	protected val humanoidNetStates = ArrayList<NetworkedElement>()
	var humanoidState by networkedEnum(HumanoidState.entries).also { humanoidNetStates.add(it) }
	var humanoidEmote by networkedEnum(HumanoidEmote.entries).also { humanoidNetStates.add(it) }
	var humanoidDance by networkedData(null, NullableBinaryStringCodec).also { humanoidNetStates.add(it) }

	protected fun inConflictingLoungeAnchor(): Boolean {
		val state = movement.anchorNetworkState

		if (state != null) {
			val entity = world.entities[state.entityID] as? LoungeableEntity

			if (entity != null) {
				val loungingIn = entity.entitiesLoungingIn(state.positionIndex)
				return loungingIn.size > 1 || this !in loungingIn
			}
		}

		return false
	}

	/**
	 * Whenever to display any armor on wearer
	 */
	open val forceNude: Boolean
		get() = false

	open val suppressUseOfTools: Boolean
		get() = false

	/**
	 * Whenever worn armor has effect on stats
	 *
	 * If false, armor will always behave like cosmetic one,
	 * no matter in which slot it is worn in
	 */
	open val disableWornArmor: Boolean
		get() = false

	protected abstract val emoteCooldownTimer: GameTimer
	protected abstract val danceTimer: GameTimer?
	protected val blinkTimer = GameTimer()
	protected abstract val blinkInterval: Vector2d

	fun setDance(dance: String?) {
		ensureIsLocal()
		this.humanoidDance = dance

		if (dance != null) {
			danceTimer?.reset(Registries.dance.getOrThrow(dance).value.duration)
		}
	}

	fun addEmote(emote: String) {
		return addEmote(HumanoidEmote.entries.valueOf(emote))
	}

	fun addEmote(emote: HumanoidEmote) {
		this.humanoidEmote = emote
		this.emoteCooldownTimer.reset()
	}

	fun requestEmote(emote: String) {
		if (emote.isNotBlank()) {
			return requestEmote(HumanoidEmote.entries.valueOf(emote))
		}
	}

	fun requestEmote(emote: HumanoidEmote) {
		if (emote != HumanoidEmote.IDLE && (humanoidEmote == HumanoidEmote.IDLE || humanoidEmote == HumanoidEmote.BLINK)) {
			addEmote(emote)
		}
	}

	override fun tickParallel(delta: Double) {
		if (isLocal) {
			if (!forceNude) {
				if (headCosmeticItem.isNotEmpty)
					effects.addEffectSources("headArmor", headCosmeticItem.effects)
				else if (headItem.isNotEmpty)
					effects.addEffectSources("headArmor", headItem.effects)

				if (chestCosmeticItem.isNotEmpty)
					effects.addEffectSources("chestArmor", chestCosmeticItem.effects)
				else if (chestItem.isNotEmpty)
					effects.addEffectSources("chestArmor", chestItem.effects)

				if (legsCosmeticItem.isNotEmpty)
					effects.addEffectSources("legsArmor", legsCosmeticItem.effects)
				else if (legsItem.isNotEmpty)
					effects.addEffectSources("legsArmor", legsItem.effects)

				if (backCosmeticItem.isNotEmpty)
					effects.addEffectSources("backArmor", backCosmeticItem.effects)
				else if (backItem.isNotEmpty)
					effects.addEffectSources("backArmor", backItem.effects)
			}

			if (!disableWornArmor) {
				val effects = ArrayList<PersistentStatusEffect>()
				effects.addAll(headItem.statusEffects)
				effects.addAll(chestItem.statusEffects)
				effects.addAll(legsItem.statusEffects)
				effects.addAll(backItem.statusEffects)
				statusController.setPersistentEffects("armor", effects)
			}

			if (!suppressUseOfTools) {
				effects.addEffectSources("primary", primaryHandItem.effects)
				effects.addEffectSources("alt", secondaryHandItem.effects)

				val effects = ArrayList<PersistentStatusEffect>()
				effects.addAll(primaryHandItem.statusEffects)
				effects.addAll(secondaryHandItem.statusEffects)
				statusController.setPersistentEffects("tools", effects)
			}

			if (emoteCooldownTimer.tick(delta))
				humanoidEmote = HumanoidEmote.IDLE
		}

		super.tickParallel(delta)

		effects.setSourcePosition("normal", movement.position)
		effects.setSourcePosition("mouth", mouthPosition)
		effects.setSourcePosition("feet", feetPosition)
		effects.setSourcePosition("headArmor", headArmorPosition)
		effects.setSourcePosition("chestArmor", chestArmorPosition)
		effects.setSourcePosition("legsArmor", legsArmorPosition)
		effects.setSourcePosition("backArmor", backArmorPosition)

		// TODO: HERE: humanoid.animate
	}

	protected fun tickTools(delta: Double, controls: Set<MoveControlType>) {
		// TODO
	}

	override fun tick(delta: Double) {
		super.tick(delta)

		if (isLocal) {
			if (blinkTimer.tick(delta)) {
				blinkTimer.reset(world.random.nextRange(blinkInterval))
				val anchor = movement.anchorState as? LoungeAnchorState

				if (humanoidEmote == HumanoidEmote.IDLE && anchor?.emote == null) {
					addEmote(HumanoidEmote.BLINK)
				}
			}
		}
	}
}
