package ru.dbotthepony.kstarbound.world.entities

import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import org.classdump.luna.Table
import org.classdump.luna.TableFactory
import ru.dbotthepony.kommons.collect.map
import ru.dbotthepony.kommons.collect.reduce
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.lua.LuaThread
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.set
import ru.dbotthepony.kstarbound.lua.setTableValue
import ru.dbotthepony.kstarbound.lua.tableOf
import ru.dbotthepony.kstarbound.lua.toByteString
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.AABBi
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.math.vector.Vector2i
import ru.dbotthepony.kstarbound.util.CarriedExecutor
import ru.dbotthepony.kstarbound.util.supplyAsync
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import java.util.PriorityQueue
import java.util.function.Supplier
import kotlin.math.PI
import kotlin.math.absoluteValue
import kotlin.math.min
import kotlin.math.roundToInt
import kotlin.math.sign
import kotlin.math.sqrt

class PathFinder(val world: World<*, *>, val start: Vector2d, val goal: Vector2d, val actorMovementParameters: ActorMovementParameters, val parameters: Parameters) : Supplier<List<PathFinder.Edge>?> {
	constructor(controller: PathController, goal: Vector2d) : this(controller.world, controller.startPosition!!, goal, controller.controller.actorMovementParameters, controller.parameters)

	// make snapshot of current things
	val chunkMap = world.chunkMap.snapshot()

	enum class Action(val isOnGround: Boolean, override val jsonName: String) : IStringSerializable {
		WALK(true, "Walk"),
		JUMP(true, "Jump"),
		ARC(false, "Arc"),
		DROP(true, "Drop"),
		SWIM(false, "Swim"),
		FLY(false, "Fly"),
		LAND(false, "Land");

		val luaName = jsonName.toByteString()!!
	}

	data class Edge(
		var cost: Double,
		var action: Action,
		var velocity: Vector2d,
		val source: Node,
		val target: Node
	) {
		fun isSame(other: Edge): Boolean {
			return action == other.action && source == other.source && target == other.target
		}
	}

	private fun defaultCostEdge(action: Action, source: Node, target: Node): Edge {
		return Edge(
			action = action,
			source = source,
			target = target,
			cost = world.geometry.diff(source.position, target.position).length,
			velocity = Vector2d.ZERO
		)
	}

	private enum class BoundingBoxType {
		FULL, DROP, STAND
	}

	private val standingBox by lazy {
		parameters.boundBox
			?: (actorMovementParameters.standingPoly?.map({ it.aabb }, { it.iterator().map { it.aabb }.reduce(AABB::combine).orNull() }) ?: AABB.withSide(Vector2d.ZERO, 0.5))
	}

	private fun boundingBox(pos: Vector2d, type: BoundingBoxType = BoundingBoxType.FULL): AABB {
		var bbox = if (type == BoundingBoxType.DROP && parameters.droppingBoundBox != null)
			parameters.droppingBoundBox
		else if (type == BoundingBoxType.STAND && parameters.standingBoundBox != null)
			parameters.standingBoundBox
		else standingBox

		bbox *= 0.99
		bbox += pos
		return bbox
	}

	private fun inLiquid(pos: Vector2d): Boolean {
		return (chunkMap.averageLiquidLevel(boundingBox(pos))?.average ?: 0.0) >= (actorMovementParameters.minimumLiquidPercentage ?: 0.5)
	}

	private fun groundCollisionRect(pos: Vector2d, type: BoundingBoxType = BoundingBoxType.FULL): AABBi {
		val bounds = AABBi.of(boundingBox(pos, type))
		return AABBi(bounds.mins, bounds.mins)
	}

	private fun onGround(pos: Vector2d, type: BoundingBoxType = BoundingBoxType.FULL): Boolean {
		val groundRect = groundCollisionRect(pos, type)

		// Check there is something under the feet.
		// We allow walking over the tops of objects (e.g. trapdoors) without being
		// able to float inside objects.
		if (chunkMap.anyCellSatisfies(boundingBox(pos, type)) { _, _, cell -> cell.foreground.material.value.collisionKind == CollisionType.DYNAMIC }) {
			// We're inside an object. Don't collide with object directly below our
			// feet:
			return chunkMap.anyCellSatisfies(groundRect) { _, _, cell -> cell.foreground.material.value.collisionKind.isFloorCollision }
		}

		// Not inside an object, allow colliding with objects below our feet:
		// We need to be for sure above platforms, but can be up to a full tile
		// below the top of solid blocks because rounded collision polys
		return chunkMap.anyCellSatisfies(groundRect) { _, _, cell -> cell.foreground.material.value.collisionKind != CollisionType.NONE } ||
			chunkMap.anyCellSatisfies(groundRect - Vector2i.POSITIVE_Y) { _, _, cell -> cell.foreground.material.value.collisionKind.isSolidNonDynamic }
	}

	private fun onSolidGround(pos: Vector2d): Boolean {
		return chunkMap.anyCellSatisfies(groundCollisionRect(pos, BoundingBoxType.DROP)) { _, _, cell -> cell.foreground.material.value.collisionKind.isSolidNonDynamic }
	}

	private fun roundToNode(pos: Vector2d): Vector2d {
		// Round pos to the nearest node.

		// Work out the distance from the entity's origin to the bottom of its
		// feet. We round Y relative to this so that we ensure we're able to
		// generate
		// paths through gaps that are *just* tall enough for the entity to fit
		// through.
		val bottom = standingBox.mins.y

		val x = (pos.x / NODE_GRANULARITY + 0.5).toInt() * NODE_GRANULARITY
		val y = ((pos.y + bottom) / NODE_GRANULARITY + 0.5).toInt() * NODE_GRANULARITY - bottom
		return Vector2d(x, y)
	}

	private fun validPosition(pos: Vector2d, type: BoundingBoxType = BoundingBoxType.FULL): Boolean {
		return !chunkMap.anyCellSatisfies(boundingBox(pos, type)) { _, _, cell ->
			cell.foreground.material.value.collisionKind.isSolidNonDynamic
		}
	}

	private data class ArcCollisionSimResult(val position: Vector2d, val collisionX: Boolean, val collisionY: Boolean)

	inner class Node(val position: Vector2d, val velocity: Vector2d? = null) : Comparable<Node> {
		var totalCost: Double = 0.0
		var parent: Edge? = null

		fun toTable(tables: TableFactory): Table {
			val table = tables.newTable(2, 0)
			table[positionIndex] = tables.from(position)
			table[velocityIndex] = tables.from(velocity)
			return table
		}

		fun store(lua: LuaThread) {
			lua.pushTable(hashSize = 2)
			lua.setTableValue("position", position)
			lua.setTableValue("velocity", velocity)
		}

		fun reconstructPath(): MutableList<Edge> {
			val result = ArrayList<Edge>()
			var parent = parent

			while (parent != null) {
				result.add(parent)
				parent = parent.source.parent
			}

			result.reverse()
			return result
		}

		override fun compareTo(other: Node): Int {
			return combinedCost.compareTo(other.combinedCost)
		}

		/**
		 * from this node to GOOOOOOOOOAAAAAAL
		 */
		val heuristicCost: Double

		init {
			val diff = world.geometry.diff(position, goal)
			heuristicCost = 2.0 * (diff.x.absoluteValue + diff.y.absoluteValue)
		}

		val combinedCost: Double
			get() = totalCost + heuristicCost

		override fun equals(other: Any?): Boolean {
			return other === this || other is Node && other.position == position && other.velocity == velocity
		}

		override fun hashCode(): Int {
			return position.hashCode() + velocity.hashCode() * 31
		}

		override fun toString(): String {
			return "Node[$position, $velocity]"
		}

		private fun simulateArcCollision(position: Vector2d, velocity: Vector2d, dt: Double): ArcCollisionSimResult {
			// Returns the new position and whether a collision in the Y axis occurred.
			// We avoid actual collision detection / resolution as that would make
			// pathfinding very expensive.

			val newPosition = position + velocity * dt

			if (validPosition(newPosition)) {
				return ArcCollisionSimResult(newPosition, false, false)
			} else {
				if (validPosition(Vector2d(newPosition.x, position.y))) {
					return ArcCollisionSimResult(Vector2d(newPosition.x, position.y), false, true)
				} else if (validPosition(Vector2d(position.x, newPosition.y))) {
					return ArcCollisionSimResult(Vector2d(position.x, newPosition.y), true, false)
				} else {
					return ArcCollisionSimResult(position, true, true)
				}
			}
		}

		// FIXME: it gives very rough estimation for directional gravity, if any at all
		@Suppress("NAME_SHADOWING")
		private fun simulateArc(callback: (target: Node, ground: Boolean) -> Unit) {
			var position = position
			var velocity = velocity!!

			var isJumping = chunkMap.gravityAt(position).dot(velocity) > 0.0 // gravity is positive
			val acceleration = acceleration

			if (acceleration == Vector2d.ZERO)
				return

			// Simulate until we're roughly NodeGranularity distance from the previous
			// node
			var rounded: Vector2d = roundToNode(position)
			val start = rounded

			while (rounded == start) {
				val speed = velocity.length
				val dt = min(0.2, if (speed != 0.0) ARC_SIMULATION_FIDELTITY / speed else sqrt(ARC_SIMULATION_FIDELTITY * 2.0 * acceleration.lengthSquared))

				val (simPosition, collidedX, collidedY) = simulateArcCollision(position, velocity, dt)
				position = simPosition
				rounded = roundToNode(simPosition)

				val upVector by lazy {
					var result = chunkMap.gravityAt(position).unitVector
					if (result == Vector2d.ZERO) result = Vector2d.POSITIVE_Y
					result
				}

				val rightVector by lazy {
					var result = chunkMap.gravityAt(position).rotate(-PI / 2.0).unitVector
					if (result == Vector2d.ZERO) result = Vector2d.POSITIVE_X
					result
				}

				if (collidedY) {
					// We've either landed or hit our head on the ceiling
					if (!isJumping) {
						// Landed
						if (velocity.dot(chunkMap.gravityAt(position).unitVector) < parameters.maxLandingVelocity) {
							callback(Node(rounded), true)
						}

						return
					} else if (onGround(rounded, BoundingBoxType.STAND)) {
						// Simultaneously hit head and landed -- this is a gap we can *just*
						// fit through. No checking of the maxLandingVelocity, since the tiles'
						// polygons are rounded, making this an easier target to hit than it
						// seems.
						callback(Node(rounded, velocity), true)
						return
					}

					// Hit ceiling. Remove y velocity
					velocity *= rightVector
				} else if (collidedX) {
					// Hit a wall, just fall down
					velocity *= upVector

					if (isJumping) {
						isJumping = false
						velocity = Vector2d.ZERO
					}
				}

				velocity += acceleration * dt

				if (isJumping && velocity.y <= 0.0) {
					// We've reached a peak in the jump and the entity can now choose to
					// change direction.

					val walkSpeed = actorMovementParameters.walkSpeed
					val runSpeed = actorMovementParameters.runSpeed
					val dot = this.velocity!!.dot(rightVector)

					if (dot != 0.0 || parameters.enableVerticalJumpAirControl) {
						val unit = velocity.unitVector

						if (walkSpeed != null && parameters.enableWalkSpeedJumps) {
							callback(Node(position, unit * walkSpeed), false)
							callback(Node(position, unit * walkSpeed * parameters.jumpDropXMultiplier), false)
						}

						if (runSpeed != null) {
							callback(Node(position, unit * runSpeed), false)
						}
					}

					// Only fall straight down if we were going straight up originally.
					// Going from an arc to falling straight down looks unnatural.

					if (dot == 0.0) {
						callback(Node(position, Vector2d.ZERO), false)
					}

					return
				}
			}

			if (!isJumping) {
				if (velocity.dot(chunkMap.gravityAt(position).unitVector) < parameters.maxLandingVelocity) {
					if (onGround(rounded, BoundingBoxType.STAND) || inLiquid(rounded)) {
						// Collision with platform
						callback(Node(rounded, velocity), true)
						return
					}
				}
			}

			check(velocity.dot(chunkMap.gravityAt(position)) != 0.0)
			callback(Node(position, velocity), false)
		}

		private fun forEachArcVelocity(velocity: Vector2d, callback: (result: Vector2d) -> Unit) {
			callback(velocity)

			val walkSpeed = actorMovementParameters.walkSpeed
			val runSpeed = actorMovementParameters.runSpeed
			val angle = -chunkMap.gravityAt(position).toAngle(Vector2d.POSITIVE_Y)

			if (parameters.enableWalkSpeedJumps && walkSpeed != null) {
				callback((Vector2d(walkSpeed) + velocity).rotate(angle))
				callback((Vector2d(-walkSpeed) + velocity).rotate(angle))
			}

			if (runSpeed != null) {
				callback((Vector2d(runSpeed) + velocity).rotate(angle))
				callback((Vector2d(-runSpeed) + velocity).rotate(angle))
			}
		}

		private fun getJumpingNeighbours(result: MutableList<Edge>) {
			val inLiquid = inLiquid(position)

			val jumpSpeed = if (inLiquid)
				// It is unknown if patching this to use liquid jump profile if applicable have undesired consequences
				actorMovementParameters.liquidJumpProfile.jumpSpeed ?: actorMovementParameters.airJumpProfile.jumpSpeed
			else
				actorMovementParameters.airJumpProfile.jumpSpeed

			if (jumpSpeed != null) {
				val jumpCost = if (inLiquid) parameters.liquidJumpCost else parameters.jumpCost

				forEachArcVelocity(chunkMap.gravityAt(position).unitVector * jumpSpeed) { velocity ->
					result.add(Edge(jumpCost, Action.JUMP, velocity, this, Node(position, velocity)))
				}

				forEachArcVelocity(chunkMap.gravityAt(position).unitVector * jumpSpeed * parameters.smallJumpMultiplier) { velocity ->
					result.add(Edge(jumpCost, Action.JUMP, velocity, this, Node(position, velocity)))
				}
			}
		}

		private fun getFlyingNeighbors(result: MutableList<Edge>) {
			val rounded = roundToNode(position)

			for (dx in -1 .. 1) {
				for (dy in -1 .. 1) {
					val pos = rounded + Vector2d(dx * NODE_GRANULARITY, dy * NODE_GRANULARITY)

					if (validPosition(pos)) {
						result.add(defaultCostEdge(Action.FLY, this, Node(pos)))
					}
				}
			}
		}

		private fun getWalkingNeighborsInDirection(result: MutableList<Edge>, direction: Double) {
			// gravity is positive vector, so it "points" up
			val angle = gravityAngle
			val up = Vector2d.POSITIVE_Y.rotate(angle)

			// rotate all vectors, so we work with local angles instead of global ones
			var forward = position + Vector2d(direction, 0.0).rotate(angle)
			val forwardAndUp = forward + up
			val forwardAndDown = forward - up

			var bounds = boundingBox(position)

			var slopeDown = false
			var slopeUp = false

			// TODO: this doesn't work with directional gravity
			val forwardGroundPos = if (direction > 0.0) Vector2d(bounds.maxs.x, bounds.mins.y) else Vector2d(bounds.mins.x, bounds.mins.y)
			val backGroundPos = if (direction > 0.0) Vector2d(bounds.mins.x, bounds.mins.y) else Vector2d(bounds.maxs.x, bounds.mins.y)

			val collisionCells = chunkMap.queryTileCollisions(groundCollisionRect(position).toDoubleAABB().padded(1.0, 1.0))

			outer@for (block in collisionCells) {
				for (side in block.poly.edges) {
					val rotSide = side.rotate(angle)
					val (p0, p1) = rotSide
					val sideDir = rotSide.direction

					val lower = if (p0.y < p1.y) p0 else p1
					val upper = if (p0.y < p1.y) p1 else p0

					if (sideDir.x != 0.0 && sideDir.y != 0.0 && (lower.y == forwardGroundPos.y || upper.y == forwardGroundPos.y)) {
						val yDir = direction * (sideDir.y / sideDir.x)

						if (world.geometry.diff(forwardGroundPos, lower).x.absoluteValue < 0.5 && yDir > 0.0) {
							slopeUp = true
							break@outer
						} else if (world.geometry.diff(backGroundPos, upper).x.absoluteValue < 0.5 && yDir < 0.0) {
							slopeDown = true
							break@outer
						}
					}
				}
			}

			// Check if it's possible to walk up a block like a ramp first
			if (slopeUp && onGround(forwardAndUp) && validPosition(forwardAndUp)) {
				// Walk up a slope
				result.add(defaultCostEdge(Action.WALK, this, Node(forwardAndUp)))
			} else if (validPosition(forward) && onGround(forward)) {
				// Walk along a flat plane
				result.add(defaultCostEdge(Action.WALK, this, Node(forward)))
			} else if (slopeDown && validPosition(forward) && validPosition(forwardAndDown) && onGround(forwardAndDown)) {
				// Walk down a slope
				result.add(defaultCostEdge(Action.WALK, this, Node(forwardAndDown)))
			} else if (validPosition(forward)) {
				// Fall off a ledge
				bounds = standingBox
				val back = if (direction > 0.0) bounds.mins.x else bounds.maxs.x
				forward -= Vector2d((1.0 - back.absoluteValue % 1.0) * direction).rotate(angle)

				val walkSpeed = actorMovementParameters.walkSpeed
				val runSpeed = actorMovementParameters.runSpeed

				if (walkSpeed != null)
					result.add(defaultCostEdge(Action.WALK, this, Node(Vector2d(walkSpeed * direction.sign).rotate(angle))))

				if (runSpeed != null)
					result.add(defaultCostEdge(Action.WALK, this, Node(Vector2d(runSpeed * direction.sign).rotate(angle))))
			}
		}

		private val acceleration: Vector2d get() {
			var gravity = chunkMap.gravityAt(position) * (actorMovementParameters.gravityMultiplier ?: 1.0)

			if (actorMovementParameters.gravityEnabled == false || actorMovementParameters.mass == 0.0) {
				gravity = Vector2d.ZERO
			}

			val buoyancy = actorMovementParameters.airBuoyancy ?: 0.0
			return -gravity * (1.0 - buoyancy)
		}

		private val gravityAngle: Double get() {
			// 0 -> points down
			// 90 -> points LEFT
			// -90 -> points RIGHT
			//        (in radians)
			return -chunkMap.gravityAt(position).toAngle(Vector2d.POSITIVE_Y)
		}

		fun computeEdges(): List<Edge> {
			val result = ArrayList<Edge>()

			if (velocity != null) {
				// Follow the current trajectory. Most of the time, this will only produce
				// one neighbor to avoid massive search space explosion, however one
				// change of X velocity is allowed at the peak of a jump.

				simulateArc { target, ground ->
					result.add(defaultCostEdge(Action.ARC, this, target))

					if (ground) {
						result.add(defaultCostEdge(Action.LAND, target, Node(target.position)))
					}
				}
			} else if (inLiquid(position)) {
				// TODO avoid damaging liquids, e.g. lava

				// We assume when we're swimming we can move freely against gravity
				getFlyingNeighbors(result)

				// Also allow jumping out of the water if we're at the surface:
				val box = AABB.withSide(position, 0.5)

				if (acceleration.y != 0.0 && (chunkMap.averageLiquidLevel(box)?.average ?: 0.0) < 1.0) {
					getJumpingNeighbours(result)
				}

				// why do they do this in original code?
				result.removeIf { !inLiquid(it.target.position) }

				for (edge in result) {
					if (edge.action == Action.FLY)
						edge.action = Action.SWIM

					edge.cost *= parameters.swimCost
				}
			} else if (acceleration.y == 0.0) {
				getFlyingNeighbors(result)
			} else if (onGround(position)) {
				getWalkingNeighborsInDirection(result, NODE_GRANULARITY)
				getWalkingNeighborsInDirection(result, -NODE_GRANULARITY)

				if (onSolidGround(position)) {
					// Add a node for dropping through a platform.
					// When that node is explored, if it's not onGround, its neighbors will
					// be falling to the ground.

					val dropPosition = position - chunkMap.gravityAt(position).unitVector

					// The physics of platforms don't allow us to drop through platforms resting
					// directly on solid surfaces. So if there is solid ground below the
					// platform, don't allow dropping through the platform:

					// TODO: KStarbound: This seems to be completely false? You can drop through platforms which rest on
					//  solid tiles just like through any other platform.

					if (!onSolidGround(dropPosition)) {
						result.add(Edge(
							parameters.dropCost,
							Action.DROP,
							Vector2d.ZERO,
							this,
							Node(dropPosition, acceleration * sqrt(2.0 / acceleration.length))
						))
					}
				}

				getJumpingNeighbours(result)
			} else {
				val position = roundToNode(position)

				// We're in the air, and can only fall now
				forEachArcVelocity(Vector2d.ZERO) { resVel ->
					Node(position, resVel).simulateArc { target, ground ->
						result.add(defaultCostEdge(Action.ARC, this, target))

						if (ground) {
							result.add(defaultCostEdge(Action.LAND, target, Node(target.position)))
						}
					}
				}
			}

			return result
		}
	}

	@JsonFactory
	data class Parameters(
		// Maximum distance from the start node to search for a path to the target
		// node
		val maxDistance: Double = 50.0,
		// If true, returns the path to the closest node to the target found, if a
		// path to the target itself could not be found.
		// Otherwise, findPath will return a None value.
		val returnBest: Boolean = false,
		// If true, end the path only on ground
		val mustEndOnGround: Boolean = false,
		// If true, allows jumps to have the entity's walk speed as horizontal
		// velocity
		val enableWalkSpeedJumps: Boolean = false,
		// if true, allows perfectly vertical jumps to change horizontal velocity at
		// the peak
		val enableVerticalJumpAirControl: Boolean = false,
		// Multiplies the cost of edges going through liquids. Can be used to
		// penalize or promote paths involving swiming.
		val swimCost: Double = 40.0,
		// The cost of jump edges.
		val jumpCost: Double = 3.0,
		// The cost of jump edges that start in liquids.
		val liquidJumpCost: Double = 10.0,
		// The cost of dropping through a platform.
		val dropCost: Double = 3.0,
		// If set, will be the default bounding box, otherwise will use
		// movementParameters.standingPoly.
		val boundBox: AABB? = null,
		// The bound box used for checking if the entity can stand at a position
		// Should be thinner than the full bound box
		val standingBoundBox: AABB? = null,
		// The bound box used for checking if the entity can drop at a position
		// Should be wider than the full bound box
		val droppingBoundBox: AABB? = null,
		// Pathing simulates jump arcs for two Y velocities: 1.0 * jumpSpeed and
		// smallJumpMultiplier * jumpSpeed. This value should be in the range
		// 0 < smallJumpMultiplier < 1.0
		val smallJumpMultiplier: Double = 0.75,
		// Mid-jump, at the peak, entities can choose to change horizontal velocity.
		// The velocities they can switch to are runSpeed, walkSpeed, and
		// (walkSpeed * jumpDropXMultiplier). The purpose of the latter option is to
		// make a vertical drop (if 0) or disable dropping (if 1). Inbetween values
		// can be used to make less angular-looking arcs.
		val jumpDropXMultiplier: Double = 0.125,
		// If provided, the following fields can be supplied to put a limit on how
		// long findPath calls can take:
		val maxFScore: Double = Double.POSITIVE_INFINITY,

		val maxNodesToSearch: Int = Int.MAX_VALUE,
		// Upper bound on the (negative) velocity that entities can land on
		// platforms
		// and ledges with. This is used to ensure there is a small amount of
		// clearance
		// over ledges to improve the scripts' chances of landing the same way we
		// simulated the jump.
		val maxLandingVelocity: Double = -5.0,
	)

	@Volatile
	private var signalStop = false

	fun signalStop() {
		signalStop = true
	}

	private var openNodes = PriorityQueue<Node>()
	private var closedNodes = HashSet<Node>()
	private var closest = Node(roundToNode(start))
	private val actualMaxNodesToSearch = parameters.maxNodesToSearch.coerceAtMost(40000) // 40000 should be more than enough for all cases

	var result = KOptional<List<Edge>?>()
		private set(value) {
			field = value
			// release memory associated with structures
			openNodes = PriorityQueue()
			closedNodes = HashSet()
			// closest = Node(roundToNode(start))
		}

	init {
		openNodes.add(closest)
		closedNodes.add(closest)
	}

	private fun canContinue(): Boolean {
		return openNodes.isNotEmpty() && (closedNodes.size - openNodes.size) < actualMaxNodesToSearch && !signalStop
	}

	fun run(maxExplore: Int): Boolean? {
		if (result.isPresent)
			return result.value != null

		val openNodes = openNodes
		val closedNodes = closedNodes
		var budget = maxExplore

		while (canContinue() && budget-- > 0) {
			val node = openNodes.remove()

			if (closest > node)
				closest = node

			if (node.position.distance(goal) <= NODE_GRANULARITY * 2.0 && (!parameters.mustEndOnGround || onGround(node.position) && node.velocity == null)) {
				closest = node
				result = KOptional(closest.reconstructPath())
				return true
			}

			for (edge in node.computeEdges()) {
				edge.target.parent = edge
				edge.target.totalCost = node.totalCost + edge.cost

				if (edge.target.combinedCost <= parameters.maxFScore && closedNodes.add(edge.target)) {
					openNodes.add(edge.target)
				}
			}
		}

		if (closedNodes.size >= 40000)
			LOGGER.warn("Path finder tried to consider more than 40000 nodes (goal $goal, closest take $closest, distance ${closest.position.distance(goal)}, onGround ${onGround(closest.position)})")

		if (!canContinue()) {
			if (parameters.returnBest) {
				result = KOptional(closest.reconstructPath())
			} else {
				result = KOptional(null as List<Edge>?)
			}
		}

		if (result.isPresent)
			return result.value != null
		else
			return null
	}

	override fun get(): List<Edge>? {
		run(Int.MAX_VALUE)
		return if (result.isPresent) result.value else null
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
		const val ARC_SIMULATION_FIDELTITY = 0.5
		const val NODE_GRANULARITY = 1.0

		private val positionIndex = "position".toByteString()!!
		private val velocityIndex = "velocity".toByteString()!!
	}
}
