package ru.dbotthepony.kstarbound.world.entities.tile

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.reflect.TypeToken
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.render.RenderLayer
import ru.dbotthepony.kstarbound.defs.InteractAction
import ru.dbotthepony.kstarbound.defs.InteractRequest
import ru.dbotthepony.kstarbound.defs.actor.PersistentStatusEffect
import ru.dbotthepony.kstarbound.defs.actor.StatModifier
import ru.dbotthepony.kstarbound.defs.`object`.LoungeOrientation
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.util.coalesceNull
import ru.dbotthepony.kstarbound.util.valueOf
import ru.dbotthepony.kstarbound.world.Direction
import ru.dbotthepony.kstarbound.world.PIXELS_IN_STARBOUND_UNIT
import ru.dbotthepony.kstarbound.world.entities.ActorEntity
import ru.dbotthepony.kstarbound.world.entities.IAnchorState
import ru.dbotthepony.kstarbound.world.entities.LoungeAnchorState
import ru.dbotthepony.kstarbound.world.entities.api.LoungeableEntity
import java.lang.Math.toRadians

class LoungeableObject(config: Registry.Entry<ObjectDefinition>) : WorldObject(config), LoungeableEntity {
	init {
		isInteractive = true
	}

	override val anchors = ArrayList<LoungeAnchorState>()

	var sitCoverImage = ""
		private set
	var sitFlipImages = false
		private set
	var sitOrientation: LoungeOrientation = LoungeOrientation.NONE
		private set

	private fun updateSitParams() {
		anchors.clear()
		val orientation = orientation ?: return

		val sitPositionsList = ArrayList<Vector2d>()

		val sitDir = orientation.directionAffinity ?: Direction.LEFT

		val sitPosition = lookupProperty("sitPosition")
		val sitPositions = lookupProperty("sitPositions")

		if (!sitPosition.isJsonNull) {
			sitPositionsList.add(vectors.fromJsonTree(sitPosition) / PIXELS_IN_STARBOUND_UNIT * sitDir.normal)
		} else if (!sitPositions.isJsonNull) {
			vectorsList.fromJsonTree(sitPositions).forEach { sitPositionsList.add(it / PIXELS_IN_STARBOUND_UNIT * sitDir.normal) }
		}

		val sitFlipDirection = lookupProperty("sitFlipDirection") { JsonPrimitive(false) }.asBoolean
		sitOrientation = LoungeOrientation.entries.valueOf(lookupProperty("sitOrientation") { JsonPrimitive("sit") }.asString)
		var sitAngle = toRadians(lookupProperty("sitAngle") { JsonPrimitive(0.0) }.asDouble)
		sitCoverImage = lookupProperty("sitCoverImage") { JsonPrimitive("") }.asString
		sitFlipImages = lookupProperty("flipImages") { JsonPrimitive(false) }.asBoolean

		if (sitDir == Direction.LEFT) {
			sitAngle *= -1f
		}

		val sitStatusEffects = ObjectArraySet<PersistentStatusEffect>()
		sitStatusEffects.addAll(statusEffects.fromJsonTree(lookupProperty("sitStatusEffects") { JsonArray() }))

		val sitEffectEmitters = ObjectArraySet<String>()
		lookupProperty("sitEffectEmitters") { JsonArray() }.asJsonArray.forEach { sitEffectEmitters.add(it.asString) }
		val sitEmote = lookupProperty("sitEmote").coalesceNull?.asString
		val sitDance = lookupProperty("sitDance").coalesceNull?.asString
		val sitArmorCosmeticOverrides = lookupProperty("sitArmorCosmeticOverrides") { JsonObject() } as JsonObject
		val sitCursorOverride = lookupProperty("sitCursorOverride").coalesceNull?.asString

		for ((i, pos) in sitPositionsList.withIndex()) {
			val anchor = LoungeAnchorState(
				position = pos,
				exitBottomPosition = pos.copy(y = yTilePosition.toDouble() + volumeBoundingBox.mins.y),
				direction = if (sitFlipDirection) sitDir.opposite else sitDir,
				angle = sitAngle,
				orientation = sitOrientation,
				// Layer all anchored entities one above the object layer, in top to bottom
				// order based on the anchor index.
				loungeRenderLayer = RenderLayer.Object.point(i.toLong()),
				controllable = false,
				emote = sitEmote,
				dance = sitDance,
				armorCosmeticOverrides = LinkedHashMap<String, JsonElement>().also { for ((k, v) in sitArmorCosmeticOverrides.entrySet()) { it[k] = v } },
				cursorOverride = sitCursorOverride,
				statusEffects = sitStatusEffects,
				effectEmitters = sitEffectEmitters
			)

			anchors.add(anchor)
		}
	}

	override fun entitiesLoungingIn(index: Int): List<ActorEntity> {
		if (index !in anchors.indices) return listOf()
		val result = ArrayList<ActorEntity>()

		world.entityIndex.iterate(metaBoundingBox, {
			if (it is ActorEntity) {
				val state = it.movement.anchorNetworkState

				if (state != null && state.entityID == entityID && state.positionIndex == index) {
					result.add(it)
				}
			}
		})

		return result
	}

	override fun entitiesLoungingIn(): List<ActorEntity> {
		val result = ArrayList<ActorEntity>()

		world.entityIndex.iterate(metaBoundingBox, {
			if (it is ActorEntity && it.movement.anchorNetworkState?.entityID == entityID) {
				result.add(it)
			}
		})

		return result
	}

	override fun parametersUpdated() {
		super.parametersUpdated()
		updateSitParams()
	}

	override fun orientationUpdated() {
		super.orientationUpdated()
		updateSitParams()
	}

	override fun interact(request: InteractRequest): InteractAction {
		val upstream = super.interact(request)

		if (upstream.type == InteractAction.Type.NONE && anchors.isNotEmpty()) {
			val interactOffset = if (direction.isRight) position - request.targetPos else request.targetPos - position
			var index = 0

			for (i in 1 until anchors.size) {
				if ((anchors[i].position + interactOffset).length < (anchors[index].position + interactOffset).length) {
					index = i
				}
			}

			return InteractAction(InteractAction.Type.SIT_DOWN, entityID, JsonPrimitive(index))
		}

		return upstream
	}

	companion object {
		private val vectors by lazy { Starbound.gson.getAdapter(Vector2d::class.java) }
		private val vectorsList by lazy { Starbound.gson.getAdapter(object : TypeToken<ImmutableList<Vector2d>>() {}) }
		private val statusEffects by lazy { Starbound.gson.getAdapter(object : TypeToken<ImmutableSet<PersistentStatusEffect>>() {}) }
	}
}
