
local string_lower = string.lower
local checkarg = checkarg

local function determineLayer(layer)
	if layer == 'foreground' then
		return false
	elseif layer == 'background' then
		return true
	else
		error('Unknown tile layer ' .. tostring(layer), 3)
	end
end

local functionsWithLayers = {
	'material',
	'mod',
	'materialHueShift',
	'modHueShift',
	'materialColor',
	'materialColorName',
	'damageTiles',
	'damageTilesPromise',
	'placeMaterial',
	'placeMaterialPromise',
	'placeMod',
	'placeModPromise',
}

for _, fnName in ipairs(functionsWithLayers) do
	local impl = assert(world[fnName])

	world[fnName] = function(pos, layer, ...)
		return impl(pos, determineLayer(layer), ...)
	end
end

local __setMaterialColorNumber = world.setMaterialColorNumber
local __setMaterialColorName = world.setMaterialColorName

function world.setMaterialColor(pos, layer, value)
	if type(value) == 'number' then
		return __setMaterialColorNumber(pos, determineLayer(layer), value)
	else
		return __setMaterialColorName(pos, determineLayer(layer), value)
	end
end

for _, fnName in ipairs({'damageTileArea', 'damageTileAreaPromise'}) do
	local impl = assert(world[fnName])

	world[fnName] = function(a, b, layer, ...)
		return impl(a, b, determineLayer(layer), ...)
	end
end

for _, name in ipairs({'entityHandItem', 'entityHandItemDescriptor'}) do
	local impl = assert(world[name])
	local fullName = 'world.' .. name

	world[name] = function(id, hand, ...)
		checkarg(hand, 2, 'string', fullName)

		local lower = string_lower(hand)

		if hand == 'primary' then
			return impl(id, true, ...)
		elseif hand == 'alt' or hand == 'secondary' then
			return impl(id, false, ...)
		else
			error('unknown tool hand ' .. hand, 2)
		end
	end
end

local entityTypes = {
	plant = 0,
	object = 1,
	vehicle = 2,
	itemDrop = 3,
	plantDrop = 4,
	projectile = 5,
	stagehand = 6,
	monster = 7,
	npc = 8,
	player = 9,
	mobile = 10,
	creature = 11,
}

local function entityTypeNamesToIntegers(input, fullName)
	if input then
		for i, v in ipairs(input) do
			-- could have used string.lower, but original engine does case-sensitive comparison here
			-- so we can save quite a lot cpu cycles for ourselves by not doing string.lower
			local lookup = entityTypes[v]

			if not lookup then
				error('invalid entity type ' .. tostring(v) .. ' for ' .. fullName .. ' in types table at index ' .. i, 3)
			end

			entityTypes[i] = lookup
		end

		return input
	end
end

local boundModes = {
	metaboundbox = 0,
	collisionarea = 1,
	position = 2
}

local function boundMode(input, fullName)
	-- collision area by default
	if not input then return 1 end

	if type(input) ~= 'string' then
		error('bad "boundMode" for ' .. fullName .. ', not a string: ' .. type(input), 3)
	end

	local lookup = boundModes[string_lower(input)]

	if not lookup then
		error('bad "boundMode": ' .. tostring(input), 3)
	end

	return lookup
end

local function order(input, fullName)
	if not input then
		return 0
	elseif input == 'random' then
		return 1
	elseif input == 'nearest' then
		return 2
	else
		error('bad "order" for ' .. fullName .. ': ' .. tostring(input), 3)
	end
end

local regularQueryFunctions = {
	['entity%sQuery'] = 'query%sImpl',
	['monster%sQuery'] = 'monsterQuery%sImpl',
	['npc%sQuery'] = 'npcQuery%sImpl',
	['itemDrop%sQuery'] = 'itemDropQuery%sImpl',
	['player%sQuery'] = 'playerQuery%sImpl',
}

for fnName, implName in pairs(regularQueryFunctions) do
	do
		local aabbImpl = assert(world[string.format(implName, 'AABB')])
		local radiusImpl = assert(world[string.format(implName, 'Radius')])
		local fnName = string.format(fnName, '')
		local fullName = 'world.' .. fnName

		world[fnName] = function(point, pointOrRadius, options)
			options = options or {}
			checkarg(point, 1, 'table', fullName, 'Vector2d')
			checkarg(options, 3, 'table', fullName)

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			if type(pointOrRadius) == 'table' then
				-- AABB
				return aabbImpl(
					{point[1], point[2], pointOrRadius[1], pointOrRadius[2]},
					entityTypeNamesToIntegers(options.includedTypes, fullName),
					options.withoutEntityId,
					boundMode(options.boundMode, fullName),
					options.callScript,
					options.callScriptArgs or {},
					options.callScriptResult,
					order(options.order, fullName),
				)
			else
				-- point + radius
				checkarg(pointOrRadius, 2, 'number', fullName, 'number or Vector2d')

				return radiusImpl(
					point,
					pointOrRadius,
					entityTypeNamesToIntegers(options.includedTypes, fullName),
					options.withoutEntityId,
					boundMode(options.boundMode, fullName),
					options.callScript,
					options.callScriptArgs or {},
					options.callScriptResult,
					order(options.order, fullName),
				)
			end
		end
	end

	do
		local lineImpl = assert(world[string.format(implName, 'Line')])
		local fnName = string.format(fnName, 'Line')
		local fullName = 'world.' .. fnName

		world[fnName] = function(p0, p1, options)
			options = options or {}
			checkarg(p0, 1, 'table', fullName, 'Vector2d')
			checkarg(p1, 2, 'table', fullName, 'Vector2d')

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			return lineImpl(
				p0, p1,
				entityTypeNamesToIntegers(options.includedTypes, fullName),
				options.withoutEntityId,
				boundMode(options.boundMode, fullName),
				options.callScript,
				options.callScriptArgs or {},
				options.callScriptResult,
				order(options.order, fullName),
			)
		end
	end

	do
		local polyImpl = assert(world[string.format(implName, 'Poly')])
		local fnName = string.format(fnName, 'Poly')
		local fullName = 'world.' .. fnName

		world[fnName] = function(poly, options)
			options = options or {}
			checkarg(poly, 1, 'table', fullName, 'Poly')

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			return polyImpl(
				poly,
				entityTypeNamesToIntegers(options.includedTypes, fullName),
				options.withoutEntityId,
				boundMode(options.boundMode, fullName),
				options.callScript,
				options.callScriptArgs or {},
				options.callScriptResult,
				order(options.order, fullName),
			)
		end
	end
end

do
	local fnName = 'object%sQuery'
	local implName = 'queryObject%sImpl'

	do
		local aabbImpl = assert(world[string.format(implName, 'AABB')])
		local radiusImpl = assert(world[string.format(implName, 'Radius')])
		local fnName = string.format(fnName, '')
		local fullName = 'world.' .. fnName

		world[fnName] = function(point, pointOrRadius, options)
			options = options or {}
			checkarg(point, 1, 'table', fullName, 'Vector2d')
			checkarg(options, 3, 'table', fullName)

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			if options.name and type(options.name) ~= 'string' then
				error('bad "name" in "options" for ' .. fullName .. ', not a table: ' .. type(options.name), 2)
			end

			if type(pointOrRadius) == 'table' then
				-- AABB
				return aabbImpl(
					options.name,
					{point[1], point[2], pointOrRadius[1], pointOrRadius[2]},
					entityTypeNamesToIntegers(options.includedTypes, fullName),
					options.withoutEntityId,
					boundMode(options.boundMode, fullName),
					options.callScript,
					options.callScriptArgs or {},
					options.callScriptResult,
					order(options.order, fullName),
				)
			else
				-- point + radius
				checkarg(pointOrRadius, 2, 'number', fullName, 'number or Vector2d')

				return radiusImpl(
					options.name,
					point,
					pointOrRadius,
					entityTypeNamesToIntegers(options.includedTypes, fullName),
					options.withoutEntityId,
					boundMode(options.boundMode, fullName),
					options.callScript,
					options.callScriptArgs or {},
					options.callScriptResult,
					order(options.order, fullName),
				)
			end
		end
	end

	do
		local lineImpl = assert(world[string.format(implName, 'Line')])
		local fnName = string.format(fnName, 'Line')
		local fullName = 'world.' .. fnName

		world[fnName] = function(p0, p1, options)
			options = options or {}
			checkarg(p0, 1, 'table', fullName, 'Vector2d')
			checkarg(p1, 2, 'table', fullName, 'Vector2d')

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			if options.name and type(options.name) ~= 'string' then
				error('bad "name" in "options" for ' .. fullName .. ', not a table: ' .. type(options.name), 2)
			end

			return lineImpl(
				options.name,
				p0, p1,
				entityTypeNamesToIntegers(options.includedTypes, fullName),
				options.withoutEntityId,
				boundMode(options.boundMode, fullName),
				options.callScript,
				options.callScriptArgs or {},
				options.callScriptResult,
				order(options.order, fullName),
			)
		end
	end

	do
		local polyImpl = assert(world[string.format(implName, 'Poly')])
		local fnName = string.format(fnName, 'Poly')
		local fullName = 'world.' .. fnName

		world[fnName] = function(poly, options)
			options = options or {}
			checkarg(poly, 1, 'table', fullName, 'Poly')

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			if options.name and type(options.name) ~= 'string' then
				error('bad "name" in "options" for ' .. fullName .. ', not a table: ' .. type(options.name), 2)
			end

			return polyImpl(
				options.name,
				poly,
				entityTypeNamesToIntegers(options.includedTypes, fullName),
				options.withoutEntityId,
				boundMode(options.boundMode, fullName),
				options.callScript,
				options.callScriptArgs or {},
				options.callScriptResult,
				order(options.order, fullName),
			)
		end
	end
end

do
	local fnName = 'loungeable%sQuery'
	local implName = 'queryLoungeable%sImpl'

	local orientations = {
		none = 0,
		sit = 1,
		lay = 2,
		stand = 3
	}

	local function orientation(input, fullName)
		if not input then return 0 end

		local lookup = orientations[input]

		if not lookup then
			if type(input) ~= 'string' then
				error('bad "orientation" in "options" for ' .. fullName .. ', not a string: ' .. type(input), 3)
			else
				error('bad "orientation" in "options" for ' .. fullName .. ', not a valid orientation: ' .. input, 3)
			end
		end

		return lookup
	end

	do
		local aabbImpl = assert(world[string.format(implName, 'AABB')])
		local radiusImpl = assert(world[string.format(implName, 'Radius')])
		local fnName = string.format(fnName, '')
		local fullName = 'world.' .. fnName

		world[fnName] = function(point, pointOrRadius, options)
			options = options or {}
			checkarg(point, 1, 'table', fullName, 'Vector2d')
			checkarg(options, 3, 'table', fullName)

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			if type(pointOrRadius) == 'table' then
				-- AABB
				return aabbImpl(
					orientation(options.orientation),
					{point[1], point[2], pointOrRadius[1], pointOrRadius[2]},
					entityTypeNamesToIntegers(options.includedTypes, fullName),
					options.withoutEntityId,
					boundMode(options.boundMode, fullName),
					options.callScript,
					options.callScriptArgs or {},
					options.callScriptResult,
					order(options.order, fullName),
				)
			else
				-- point + radius
				checkarg(pointOrRadius, 2, 'number', fullName, 'number or Vector2d')

				return radiusImpl(
					orientation(options.orientation),
					point,
					pointOrRadius,
					entityTypeNamesToIntegers(options.includedTypes, fullName),
					options.withoutEntityId,
					boundMode(options.boundMode, fullName),
					options.callScript,
					options.callScriptArgs or {},
					options.callScriptResult,
					order(options.order, fullName),
				)
			end
		end
	end

	do
		local lineImpl = assert(world[string.format(implName, 'Line')])
		local fnName = string.format(fnName, 'Line')
		local fullName = 'world.' .. fnName

		world[fnName] = function(p0, p1, options)
			options = options or {}
			checkarg(p0, 1, 'table', fullName, 'Vector2d')
			checkarg(p1, 2, 'table', fullName, 'Vector2d')

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			return lineImpl(
				orientation(options.orientation),
				p0, p1,
				entityTypeNamesToIntegers(options.includedTypes, fullName),
				options.withoutEntityId,
				boundMode(options.boundMode, fullName),
				options.callScript,
				options.callScriptArgs or {},
				options.callScriptResult,
				order(options.order, fullName),
			)
		end
	end

	do
		local polyImpl = assert(world[string.format(implName, 'Poly')])
		local fnName = string.format(fnName, 'Poly')
		local fullName = 'world.' .. fnName

		world[fnName] = function(poly, options)
			options = options or {}
			checkarg(poly, 1, 'table', fullName, 'Poly')

			if options.callScriptArgs and type(options.callScriptArgs) ~= 'table' then
				error('bad "callScriptArgs" in "options" for ' .. fullName .. ', not a table: ' .. type(options.callScriptArgs), 2)
			end

			return polyImpl(
				orientation(options.orientation),
				poly,
				entityTypeNamesToIntegers(options.includedTypes, fullName),
				options.withoutEntityId,
				boundMode(options.boundMode, fullName),
				options.callScript,
				options.callScriptArgs or {},
				options.callScriptResult,
				order(options.order, fullName),
			)
		end
	end
end
